from itertools import product
from typing import List, Tuple, Dict, Set, Iterator, Iterable, Union, Sequence
from math import sqrt
import numpy as np
from scipy.spatial.distance import cdist
from ase import Atoms
from clease import tools
from clease.datastructures import FourVector, Figure
from .cluster import Cluster
from .cluster_fingerprint import ClusterFingerprint

__all__ = ("ClusterGenerator", "SitesWithinCutoff")


class ClusterGenerator:
    """
    Class for generating cluster info that is independent of the
    template
    """

    def __init__(self, prim_cell: Atoms) -> None:
        self.with_cutoff = SitesWithinCutoff(self)
        self.prim = prim_cell

    @property
    def prim(self) -> Atoms:
        return self._prim

    @prim.setter
    def prim(self, other: Atoms) -> None:
        if not isinstance(other, Atoms):
            raise TypeError("prim must be an Atoms object")
        self._prim = other
        # Get the transposed cell as a contiguous array
        self.prim_cell_T = np.ascontiguousarray(other.get_cell().T)
        self.prim_cell_invT = np.linalg.inv(self.prim_cell_T)

    def __eq__(self, other) -> bool:
        if not isinstance(other, ClusterGenerator):
            return NotImplemented

        return np.allclose(self.prim_cell_invT, other.prim_cell_invT) and self.prim == other.prim

    def _fv_to_cart(self, fv: FourVector) -> np.ndarray:
        """Helper function to convert a FourVector object to cartesian using the internal
        primitive cell."""
        return fv.to_cartesian(self.prim, transposed_cell=self.prim_cell_T)

    def _as_euc(self, x: Union[np.ndarray, FourVector]) -> np.ndarray:
        """Helper function to translate vector to NumPy array format"""
        if isinstance(x, FourVector):
            return self._fv_to_cart(x)
        # Assume it's already in cartesian coordinates
        return x

    def many_to_four_vector(
        self, cartesian: np.ndarray, sublattices: Sequence[int]
    ) -> List[FourVector]:
        """Translate many positions into FourVector's"""

        if cartesian.ndim != 2:
            raise ValueError(f"Cartesian must be 2 dimensional, got {cartesian.ndim:d}")

        if cartesian.shape[1] != 3:
            raise ValueError(
                "Cartesian should have a length of 3 on the 2nd dimension, "
                f"got {cartesian.shape[1]:d}."
            )

        if len(sublattices) != len(cartesian):
            raise ValueError(
                "Number of sublattices should be same as number of cartesian positions"
            )

        delta = cartesian - self.prim.positions[sublattices, :]
        inv = self.prim_cell_invT.dot(delta.T).T
        # Round off, and cast to integer
        ints = np.rint(inv).astype(int)  # dtype: np.int64

        # One position per sublattice
        assert len(sublattices) == len(ints)
        # Iterating the array as a list is more efficient than iterating the numpy array.
        return [FourVector(*vals, subl) for vals, subl in zip(ints.tolist(), sublattices)]

    def to_four_vector(self, cartesian: np.ndarray, sublattice: int = None) -> FourVector:
        """Translate a position in Cartesian coordinates to its FourVector"""
        if cartesian.ndim != 1:
            raise ValueError(f"Cartesian positions must be 1-dimensional, got {cartesian.ndim:d}")

        if sublattice is None:
            sublattice = self.get_lattice(cartesian)

        return self.many_to_four_vector(cartesian[None, :], [sublattice])[0]

    def to_cartesian(self, *four_vectors: FourVector) -> np.ndarray:
        """Convert one or more FourVectors into their Cartesian coordinates."""
        # We calculate the dot product here, instead of using the fv.to_cartesian, in order
        # to better utilize the NumPy vectorization.
        # So first we allocate the necessary arrays
        X = np.zeros((len(four_vectors), 3), dtype=int)
        for ii, fv in enumerate(four_vectors):
            X[ii, :] = fv.ix, fv.iy, fv.iz
        subl = [fv.sublattice for fv in four_vectors]
        return self.prim_cell_T.dot(X.T).T + self.prim.positions[subl]

    def get_lattice(self, pos: np.ndarray) -> int:
        """
        Return the corresponding sublattice of a cartesian position

        :param pos: array of length 3 Cartesian position
        """
        shifts = self.prim.get_positions()
        reshaped = np.reshape(pos, (1, 3))
        wrapped = tools.wrap_positions_3d(
            reshaped, self.prim.get_cell(), cell_T_inv=self.prim_cell_invT
        )
        diff_sq = np.sum((wrapped[0, :] - shifts) ** 2, axis=1)
        return int(np.argmin(diff_sq))

    def eucledian_distance(
        self,
        x1: Union[np.ndarray, FourVector],
        x2: Union[np.ndarray, FourVector],
    ) -> float:
        """
        Eucledian distance between to FourVectors in cartesian coordinates.
        Either the arrays are given as a FourVector object, in which case they are translated
        into the cartesian coordiates, or they are assumed to already be cartesian coordinates.
        This allows for either passing in a pre-calculated euclidian vector,
        or the FourVector itself.

        :param x1: First vector
        :param x2: Second vector
        """

        # Find the distance vector
        euc1 = self._as_euc(x1)
        euc2 = self._as_euc(x2)
        return _1d_array_norm(euc2 - euc1)

    @property
    def shortest_diag(self) -> float:
        cellT = self.prim_cell_T

        def _iter_diags() -> Iterable[float]:
            """Helper function to iterate all diagonals"""
            for w in product([-1, 0, 1], repeat=3):
                if all(x == 0 for x in w):
                    continue

                diag: np.ndarray = cellT.dot(w)
                length: float = _1d_array_norm(diag)
                yield length

        # Find the shortest length
        shortest = min(_iter_diags())
        return shortest

    @property
    def num_sub_lattices(self) -> int:
        return len(self.prim)

    def sites_within_cutoff(self, cutoff: float, x0: FourVector) -> Iterable[FourVector]:
        min_diag = self.shortest_diag
        max_int = int(cutoff / min_diag) + 1

        # Pre-calculated cartesian representation
        cart0 = self._fv_to_cart(x0)

        def filter_func(fv: FourVector) -> bool:
            d = self.eucledian_distance(cart0, fv)
            return 1e-5 < d < cutoff

        all_sites = product(
            range(-max_int, max_int + 1),
            range(-max_int, max_int + 1),
            range(-max_int, max_int + 1),
            range(self.num_sub_lattices),
        )

        return filter(filter_func, (FourVector(*site) for site in all_sites))

    def get_fp(self, figure: Figure) -> ClusterFingerprint:
        """Generate finger print for a Figure object.

        Args:
            figure (Figure): The figure to find the finger print from.

        Returns:
            ClusterFingerprint: The finger print of the cluster.
        """
        X = figure.to_cartesian(self.prim, transposed_cell=self.prim_cell_T)
        return positions_to_fingerprint(X)

    def prepare_within_cutoff(
        self, cutoff: float, lattice: int
    ) -> Dict[FourVector, Set[FourVector]]:
        """Prepare all sites which are within the cutoff sphere. Note, this only prepares sites
        which are pair-wise within the cutoff, and does not consider the distance to the
        center-of-mass. This needs to be checked for the individual figure which is created from
        these sites."""

        within_cutoff = {}
        x0 = FourVector(0, 0, 0, lattice)
        sites = set(self.with_cutoff.get(cutoff, lattice))
        within_cutoff[x0] = sites

        dist_fnc = self.eucledian_distance  # Cache the distance function
        for s in sites:
            # Precalculate the cartesian representation of the FourVector
            cart_s = self._fv_to_cart(s)
            nearby = set(s1 for s1 in sites if s1 != s and dist_fnc(s1, cart_s) <= cutoff)

            within_cutoff[s] = nearby
        return within_cutoff

    def figure_iterator(self, size: int, cutoff: float, ref_lattice: int) -> Iterator[Figure]:
        """Iterate all possible figures of a given size, cutoff radius, and reference lattice.
        All figures are guaranteed to have a radius smaller than the cutoff radius.

        Args:
            size (int): Size of the clusters to be generated, e.g. 2 for 2-body clusters.
            cutoff (float): Cutoff sphere from the center of mass of the cluster.
            ref_lattice (int, optional): Site which clusters are generated from. Any cluster
            always contains this site.

        Yields:
            Iterator[Figure]: Iterable with all acceptable figures.
        """

        def is_figure_ok(figure: Figure) -> bool:
            """Does the figure obey the cutoff radius?
            The diameter can be larger than the maximum distance between two atoms
            in the cluster, and so the site_iterator can prepare clusters with too large
            diameters.

            Returns:
                bool: The diameter of the cluster is within the cutoff
            """
            return figure.get_diameter(self.prim, transposed_cell=self.prim_cell_T) < cutoff

        cutoff_lut = self.prepare_within_cutoff(cutoff, ref_lattice)
        return filter(is_figure_ok, site_iterator(cutoff_lut, size, ref_lattice))

    def generate(
        self, size: int, cutoff: float, ref_lattice: int
    ) -> Tuple[List[List[Figure]], List[ClusterFingerprint]]:
        """Generate all possible figures of a given size, are within a given cutoff radius
        (from the center of mass of the figure), and from a reference lattice.

        Args:
            size (int): Size of the clusters to be generated, e.g. 2 for 2-body clusters.
            cutoff (float): Cutoff sphere from the center of mass of the cluster.
            ref_lattice (int, optional): Site which clusters are generated from. Any cluster
            always contains this site.

        Returns:
            List[List[Figure]], List[ClusterFingerprint]: The collection of figures and
                their corresponding fingerprints.
        """
        clusters: List[List[Figure]] = []
        all_fps: List[ClusterFingerprint] = []

        for new_figure in self.figure_iterator(size, cutoff, ref_lattice):
            # The entries in the figure must be ordered, due to equiv_sites
            # TODO: This ordereding should not be necessary.
            new_figure = self._order_by_internal_distances(new_figure)

            fp = self.get_fp(new_figure)

            # Find the group
            try:
                group = all_fps.index(fp)
                clusters[group].append(new_figure)
            except ValueError:
                # Does not exist, create a new group
                clusters.append([new_figure])
                all_fps.append(fp)

        return clusters, all_fps

    def _get_internal_distances(self, figure: Figure, sort: bool = False) -> np.ndarray:
        """
        Return all the internal distances of a figure.

        If ``sort`` is True, then distances are sorted from
        largest->smallest distance for each component.

        Returns a 2D array with the internal distances. If the array is *unsorted*,
            then entry ``(i, j)`` is the eucledian distance between figures ``i`` and ``j``.
        """
        comp = figure.components

        # Convert each FourVector component to cartesian coordinates
        x = np.array([self._fv_to_cart(c) for c in comp], dtype=float)
        # Build the distance array, such that elements (i, j) is the
        # Euclidean distance betweens elements i and j.
        arr = cdist(x, x, metric="euclid")
        # Round to 6 digits for numerical stability.
        arr = np.round(arr, 6)
        if sort:
            # fliplr for reversing the sort. Sorted such that each row has largest->smallest
            # distance (i.e. like reverse=True does in Python's sort() method).
            arr = np.fliplr(np.sort(arr, axis=1))
        return arr

    def _order_by_internal_distances(self, figure: Figure) -> Figure:
        """Order the Figure by internal distances. Returns a new instance of the Figure."""
        dists = self._get_internal_distances(figure, sort=True)
        fvs = figure.components

        # Sort according to the figure with the largest internal distances, lexographically.
        # Largest internal distances first.
        order = np.lexsort(dists.T)[::-1]
        return Figure((fvs[i] for i in order))

    def to_atom_index(self, cluster: Cluster, lut: Dict[FourVector, int]) -> List[List[int]]:
        """
        Convert the integer vector representation to an atomic index

        :param clusters: List of clusters (in integer vector representation)

        :param template: Template atoms to use

        :param lut: Look up table for 4-vectors to indices
        """
        # pylint: disable=no-self-use
        return [[lut[fv] for fv in fig.components] for fig in cluster.figures]

    def equivalent_sites(self, figure: Figure) -> List[List[int]]:
        """Find the equivalent sites of a figure."""
        dists = self._get_internal_distances(figure, sort=True)
        equiv_sites = []
        for i, d1 in enumerate(dists):
            for j in range(i + 1, len(dists)):
                d2 = dists[j]
                if np.allclose(d1, d2):
                    equiv_sites.append((i, j))

        # Merge pairs into groups
        merged: List[Set[int]] = []
        for equiv in equiv_sites:
            found_group = False
            for m in merged:
                if any(x in m for x in equiv):
                    m.update(equiv)
                    found_group = True
            if not found_group:
                merged.append(set(equiv))
        return [list(x) for x in merged]

    def get_max_distance(self, figure: Figure):
        """Return the maximum distance of a figure."""
        internal_dists = self._get_internal_distances(figure, sort=False)
        return internal_dists.max()


class SitesWithinCutoff:
    """
    Class that keeps track of nearby sites. It automatically re-calculates if
    the cutoff changes. If it can re-use a previous result it will do that.
    """

    def __init__(self, generator: ClusterGenerator) -> None:
        self.generator = generator
        self.cutoff = 0.0
        # Dictionary mapping a ref lattice to a list of four-vectors
        self.pre_calc: Dict[int, List[FourVector]] = {}

    def must_generate(self, cutoff: float, ref_lattice: int) -> bool:
        """
        Check if we can re-use a precalculated result
        """
        if abs(cutoff - self.cutoff) > 1e-10:
            return True
        return ref_lattice not in self.pre_calc

    def get(self, cutoff: float, ref_lattice: int) -> List[FourVector]:
        """
        Return sites within the cutoff
        """
        if self.must_generate(cutoff, ref_lattice):
            sites = self.generator.sites_within_cutoff(cutoff, FourVector(0, 0, 0, ref_lattice))
            self.pre_calc[ref_lattice] = list(sites)
            self.cutoff = cutoff
        return self.pre_calc[ref_lattice]


def site_iterator(
    within_cutoff: Dict[FourVector, Set[FourVector]], size: int, ref_lattice: int
) -> Iterator[Figure]:
    """
    Return an iterator of all combinations of sites within a cutoff

    :param within_cutoff: Dictionary returned by `prepare_within_cutoff`

    :param size: Cluster size

    :param ref_lattice: Reference lattice
    """
    x0 = FourVector(0, 0, 0, ref_lattice)

    def recursive_yield(remaining, current):
        if len(current) == size - 1:
            # This is the full Figure object
            components = [x0] + current
            fig = Figure(components)
            yield fig
        else:
            for next_item in remaining:
                # Avoid double counting
                if any(next_item <= v for v in current):
                    continue
                rem = set.intersection(remaining, within_cutoff[next_item])
                yield from recursive_yield(rem, current + [next_item])

    for v in within_cutoff[x0]:
        rem = within_cutoff[x0].intersection(within_cutoff[v])
        yield from recursive_yield(rem, [v])


def positions_to_fingerprint(X: np.ndarray) -> ClusterFingerprint:
    """
    Generate finger print given a position matrix
    """
    # Center around the center of mass
    X = np.array(X)
    com = np.mean(X, axis=0)
    X -= com
    X = X.dot(X.T)
    # The positions squared
    diag_positions = np.diagonal(X)

    N = X.shape[0]
    # Put all the upper triangle off-diagonals into a 1D array one after another,
    # [off_diag1, off_diag2, ...]
    off_diag_len = N * (N - 1) // 2
    off_diag = np.zeros(off_diag_len)
    offset = 0
    for i in range(1, N):
        diag = np.diagonal(X, offset=i)
        off_diag[offset : (len(diag) + offset)] = diag
        offset += len(diag)
    # Sort positions, such that the largest position is first
    diag_positions = np.sort(diag_positions)[::-1]
    off_diag = np.sort(off_diag)
    # Append the off-diagonal elements after the diagonal, into a 1D array
    inner = np.append(diag_positions, off_diag)
    return ClusterFingerprint(fp=inner)


def _1d_array_norm(arr: np.ndarray) -> float:
    """Calculate the norm of a 1-d vector.
    Assumes the vector is 1-dimensional and real, and will not be checked.

    More efficient norm than ``np.linalg.norm`` for small vectors,
    since we already know the dimensionality, all numbers are real,
    and that the result is scalar.
    """
    return sqrt(arr.dot(arr))



from typing import List, Tuple
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.figure import Figure
from clease import Evaluate, ConvexHull
from clease.evaluate import supports_alpha_cv


def plot_fit(evaluate: Evaluate, plot_args: dict = None, interactive: bool = False) -> Figure:
    """
    Figure object calculated (DFT) and predicted energies.
    If the plot_args dictionary contains keys,
    return  figure object to relate plot_args  keys

    :param evaluate: Use the evaluate object to define the plot argument.
    :param plot_args: plot_args dictionary contains:

        - "xlabel": x-axis label
        - "ylabel": y-axis label
        - "title": title of plot

    :param interactive: Add interactive elements to the plot.

    :return: Figure instance of plot
    """
    if plot_args is None:
        plot_args = {}
    X = evaluate.get_energy_true()
    Y = evaluate.get_energy_predict()
    xlabel = plot_args.get("xlabel", r"E$_{DFT}$ (eV/atom)")
    ylabel = plot_args.get("ylabel", r"E$_{CE}$ (eV/atom)")
    title = plot_args.get("title", f"Fit using {len(evaluate.e_dft)} data points.")

    # rmin, rmax set the plot range of x, y coordinate
    if np.size(X) and np.size(Y) != 0:
        e_range = max(np.append(X, Y)) - min(np.append(X, Y))
        rmin = min(np.append(X, Y)) - 0.05 * e_range
        rmax = max(np.append(X, Y)) + 0.05 * e_range
    else:
        rmin = -10
        rmax = 10
    linear_fit = np.arange(rmin - 10, rmax + 10, 1)
    cv_name = evaluate.scoring_scheme.lower()

    if cv_name == "k-fold":
        # Figure out the k
        nsplits = evaluate.nsplits
        cv_name = f"{nsplits}-fold CV"

    cv = evaluate.get_cv_score() * 1000
    rmse = evaluate.rmse() * 1000

    fig = plt.figure()
    ax = fig.add_subplot(111)
    ax.set_title(title)
    ax.set_ylabel(ylabel)
    ax.set_xlabel(xlabel)
    ax.axis([rmin, rmax, rmin, rmax])
    ax.text(
        0.95,
        0.01,
        cv_name + f" = {cv:.3f} meV/atom\n" f"RMSE = {rmse:.3f} meV/atom",
        verticalalignment="bottom",
        horizontalalignment="right",
        transform=ax.transAxes,
        fontsize=12,
    )
    ax.plot(linear_fit, linear_fit, "r")
    plot_line = ax.plot(X, Y, "bo", mfc="none")[0]

    if interactive:
        # pylint: disable=import-outside-toplevel
        from clease.interactive_plot import ShowStructureOnClick, AnnotatedAx

        annotations = _make_annotations_plot_fit(evaluate)
        # Construct the annotated axis objects.
        annotated_ax = AnnotatedAx(
            ax,
            [plot_line],
            annotations,
            structure_names=[evaluate.names],
        )

        # Attach interactivity to the fig object.
        ShowStructureOnClick(fig, annotated_ax, evaluate.settings.db_name)

    return fig


def plot_fit_residual(
    evaluate: Evaluate, plot_args: dict = None, interactive: bool = False
) -> Figure:
    """
    Figure object subtracted (DFT) and predicted energies.
    If the plot_args dictionary contains keys,
    return  figure object to relate plot_args  keys

    :param evaluate: Use the evaluate object
        to define the plot argument.
    :param plot_args: plot_args dictionary contains:

        - "xlabel": x-axis label
        - "ylabel": y-axis label
        - "title": title of plot
    :param interactive: Add interactive elements to the plot.

    :return: Figure instance of plot
    """
    if plot_args is None:
        plot_args = {}
    X = evaluate.get_energy_true()
    Y = evaluate.get_energy_predict() - X  # eV/atom
    Y *= 1000  # meV/atom
    xlabel = plot_args.get("xlabel", "#OCC")
    ylabel = plot_args.get("ylabel", r"$E_{DFT} - E_{pred}$ (meV/atom)")
    title = plot_args.get("title", "Residual (v)")

    gridspec_kw = {"wspace": 0.0, "width_ratios": [5, 1]}
    fig, ax = plt.subplots(ncols=2, sharey=True, gridspec_kw=gridspec_kw)
    ax[0].axhline(0, ls="--")
    plot_line = ax[0].plot(X, Y, "v", mfc="none")[0]
    ax[0].set_xlabel(r"$E_{DFT}$ (eV/atom)")
    ax[0].set_ylabel(ylabel)
    ax[0].set_title(title)

    hist, bin_edges = np.histogram(Y, bins=30)
    h = bin_edges[1] - bin_edges[0]
    ax[1].barh(bin_edges[:-1], hist, height=h, color="#bdbdbd")
    ax[1].set_xlabel(xlabel)
    ax[1].spines["right"].set_visible(False)
    ax[1].spines["top"].set_visible(False)

    if interactive:
        # pylint: disable=import-outside-toplevel
        from clease.interactive_plot import ShowStructureOnClick, AnnotatedAx

        annotations = _make_annotations_plot_fit(evaluate)
        # Construct the annotated axis objects.
        annotated_ax = AnnotatedAx(
            ax[0],
            [plot_line],
            annotations,
            structure_names=[evaluate.names],
        )

        # Attach interactivity to the fig object.
        ShowStructureOnClick(fig, annotated_ax, evaluate.settings.db_name)

    return fig


def plot_eci(
    evaluate: Evaluate,
    plot_args: dict = None,
    ignore_sizes=(),
    interactive: bool = False,
) -> Figure:
    """
    Figure object of ECI value according to cluster diameter
    If the plot_args dictionary contains keys,
    return  figure object to relate plot_args  keys

    :param evaluate: Use the evaluate object
        to define the plot argument.
    :param plot_args: plot_args dictionary contains:

        - "xlabel": x-axis label
        - "ylabel": y-axis label
        - "title": title of plot
        - "sizes": list of int to include n-body cluster in plot
    :param ignore_sizes: list of ints
        Sizes listed in this list will not be plotted.
        E.g. ``ignore_sizes=[0]`` will exclude the 0-body cluster.
        Default is to not ignore any clusters.
    :param interactive: Add interactive elements to the plot.

    :return: Figure instance of plot
    """
    if plot_args is None:
        plot_args = {}
    # eci_by_size dictionary contains eci, name, distance
    eci_by_size = evaluate.get_eci_by_size()
    xlabel = plot_args.get("xlabel", r"Cluster diameter ($n^{th}$ nearest neighbor)")
    ylabel = plot_args.get("ylabel", "ECI (eV/atom)")
    title = plot_args.get("title", "Plot ECI")
    sizes = plot_args.get("sizes", list(eci_by_size.keys()))

    markers = ["o", "v", "x", "D", "^", "h", "s", "p"]
    lines = []
    annotations = []

    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1)
    ax.axhline(0.0, ls="--", color="grey")
    ax.xaxis.set_major_locator(plt.MultipleLocator(1))
    ax.set_title(title)
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)

    for size in sizes:
        if ignore_sizes and size in ignore_sizes:
            continue
        data = eci_by_size[size]
        # Add 1, as NN starts from 1 and not 0
        X = np.array(data["distance"]) + 1
        Y = data["eci"]
        mrk = markers[size % len(markers)]
        line = ax.plot(X, Y, label=f"{size}-body", marker=mrk, mfc="none", ls="", markersize=8)

        # Make annotations for interactive plots, since we have all the data we need prepared here.
        if interactive:
            lines.append(line[0])
            annot = [
                (
                    f"Size: {size}\nDiameter: {dist:d}\nName: {name}\n"
                    f"Radius: {radius:.3f} Å\nECI: {eci:.4f} eV/atom"
                )
                for dist, name, eci, radius in zip(X, data["name"], Y, data["radius"])
            ]
            annotations.append(annot)
    ax.legend()

    if interactive:
        # pylint: disable=import-outside-toplevel
        from clease.interactive_plot import InteractivePlot, AnnotatedAx

        # Construct the annotated axis objects.
        annotated_ax = AnnotatedAx(
            ax,
            lines,
            annotations,
        )

        # Attach interactivity to the fig object.
        InteractivePlot(fig, annotated_ax)

    return fig


def plot_cv(evaluate: Evaluate, plot_args: dict = None) -> Figure:
    """
    Figure object of CV values according to alpha values
    If the plot_args dictionary contains keys,
    return  figure object to relate plot_args  keys

    :param evaluate: Use the evaluate object
        to define the plot argument.
    :param plot_args: plot_args dictionary contains:

        - "xlabel": x-axis label
        - "ylabel": y-axis label
        - "title": title of plot

    :return: Figure instance of plot
    """
    if not supports_alpha_cv(evaluate.scheme):
        raise ValueError(f"Scheme {evaluate.scheme!r} doesn't support alpha CV.")
    if plot_args is None:
        plot_args = {}
    alpha_cv_data = evaluate.cv_scores
    xlabel = plot_args.get("xlabel", "alpha")
    ylabel = plot_args.get("ylabel", "CV score (meV/atom)")
    title = plot_args.get("title", "CV score vs. alpha")
    fig = plt.figure()
    ax = fig.add_subplot(111)
    ax.set_title(title)
    ax.set_ylabel(ylabel)
    ax.set_xlabel(xlabel)

    # this variable is for getting minimum cv
    min_cv = min(alpha_cv_data, key=lambda x: x["cv"])
    X = []
    Y = []
    for data in alpha_cv_data:
        X.append(data["alpha"])
        # Convert from eV/atom to meV/atom
        Y.append(data["cv"] * 1000)
    ax.plot(X, Y)
    ax.text(
        0.65,
        0.01,
        f"min. CV score:\n"
        f"alpha = {min_cv['alpha']:.10f} \n"
        f"CV = {min_cv['cv'] * 1000.0:.3f}"
        f" meV/atom",
        verticalalignment="bottom",
        horizontalalignment="left",
        transform=ax.transAxes,
        fontsize=10,
    )
    # Display the point with the best CV score as a red circle
    idx = np.argmin(Y)
    ax.plot(X[idx], Y[idx], "ro", markersize=5)
    return fig


def plot_convex_hull(evaluate: Evaluate, interactive: bool = False) -> Figure:
    """Plot the convex hull of an evaluate object.

    Args:
        evaluate (Evaluate): The Evaluate object to draw the convex hull from.
        interactive (bool, optional): Plot as an interactive figure?. Defaults to False.
    """
    e_pred = evaluate.get_energy_predict()

    cnv_hull = ConvexHull(evaluate.settings.db_name, select_cond=evaluate.select_cond)

    # Construct the figure object, and plot the DFT points.
    # Also fetch the Line2D objects which are drawn on the figure (i.e., the scatter points)
    # dft_lines is for interactive plotting.
    fig, dft_lines = cnv_hull.plot(return_lines=True)

    # `conc_per_frame` is the concentration with respect to the total number of atoms for
    # each frame
    conc_per_frame = evaluate.atomic_concentrations

    # `concs` is dictionary with the keys of species with value being the
    # concentrations among the frames
    # pylint: disable=protected-access
    concs = {key: [] for key in cnv_hull._unique_elem}
    for frame_conc in conc_per_frame:
        for key, value in concs.items():
            value.append(frame_conc.get(key, 0.0))

    form_en = [cnv_hull.get_formation_energy(c, e) for c, e in zip(conc_per_frame, e_pred.tolist())]
    # Draw the CE energies, and fetch the lines, in case we need them for interactive plotting.
    _, ce_lines = cnv_hull.plot(
        fig=fig,
        concs=concs,
        energies=form_en,
        marker="x",
        return_lines=True,
    )

    fig.suptitle("Convex hull DFT (o), CE (x)")

    if interactive:
        # pylint: disable=import-outside-toplevel
        from clease.interactive_plot import ShowStructureOnClick, AnnotatedAx

        ax_list = fig.get_axes()

        annotations = _make_annotations_hull(evaluate)
        # Construct the annotated axis objects.
        annotated_axes = []
        for ii, ax in enumerate(ax_list):
            data_points = (dft_lines[ii], ce_lines[ii])
            annotated_ax = AnnotatedAx(
                ax,
                data_points,
                annotations,
                structure_names=[evaluate.names, evaluate.names],
            )
            annotated_axes.append(annotated_ax)

        ShowStructureOnClick(fig, annotated_axes, evaluate.settings.db_name)

    return fig


def _make_annotations_hull(evaluate: Evaluate) -> Tuple[List[str], List[str]]:
    """Helper function to make annotations for interactive plots."""
    e_pred = evaluate.get_energy_predict()
    e_dft = evaluate.get_energy_true()

    def format_annotation_dft(idx):
        name = evaluate.names[idx]
        row_id = evaluate.row_ids[idx]
        en = e_dft[idx]
        return f"DB ID: {row_id}\nName: {name}\nE(DFT): {en:.4f} eV/atom"

    def format_annotation_ce(idx):
        name = evaluate.names[idx]
        row_id = evaluate.row_ids[idx]
        e_ce = e_pred[idx]
        return f"DB ID: {row_id}\nName: {name}\nE(CE): {e_ce:.4f} eV/atom"

    N = len(evaluate.names)
    an_dft = [format_annotation_dft(idx) for idx in range(N)]
    an_ce = [format_annotation_ce(idx) for idx in range(N)]
    return an_dft, an_ce


def _make_annotations_plot_fit(evaluate: Evaluate) -> Tuple[List[str], List[str]]:
    """Helper function to make annotations for interactive plots."""
    e_pred = evaluate.get_energy_predict()
    e_dft = evaluate.get_energy_true()

    def format_annotation(idx):
        name = evaluate.names[idx]
        row_id = evaluate.row_ids[idx]
        en = e_dft[idx]
        e_ce = e_pred[idx]
        return f"DB ID: {row_id}\nName: {name}\nE(DFT): {en:.4f} eV/atom\nE(CE): {e_ce:.4f} eV/atom"

    N = len(e_pred)
    return ([format_annotation(idx) for idx in range(N)],)



import numpy as np
import attr
from clease.calculator import Clease
from clease.montecarlo.averager import Averager
from clease.datastructures.mc_step import MCStep
from .mc_observer import MCObserver


@attr.define(slots=True)
class SGCQuantities:
    """Container for observed quantities in the SGCObserver"""

    energy: Averager = attr.field()
    energy_sq: Averager = attr.field()
    singlets: np.ndarray = attr.field()
    singlets_sq: np.ndarray = attr.field()
    singl_eng: np.ndarray = attr.field()
    counter: int = attr.field(default=0)

    def reset(self) -> None:
        self.counter = 0
        self.energy.clear()
        self.energy_sq.clear()
        self.singlets[:] = 0.0
        self.singlets_sq[:] = 0.0
        self.singl_eng[:] = 0.0


class SGCObserver(MCObserver):
    """
    Observer mainly intended to track additional quantities needed when
    running SGC Monte Carlo. This observer has to be executed on every MC step.

    Parameters:

    calc: `clease.calculators.Clease`
        Clease calculator

    observe_singlets: bool
        Whether the singlet values of the calculator are measured during each observation.
        Measuring singlets is slightly more expensive, so this is disabled by default.
    """

    name = "SGCObersver"

    def __init__(self, calc: Clease, observe_singlets: bool = False):
        self.observe_singlets = observe_singlets
        super().__init__()
        self.calc = calc
        initial_energy = calc.get_potential_energy()
        n_singlets = len(self.calc.get_singlets())
        self.quantities = SGCQuantities(
            energy=Averager(ref_value=initial_energy),
            energy_sq=Averager(ref_value=initial_energy**2),
            singlets=np.zeros(n_singlets, dtype=np.float64),
            singlets_sq=np.zeros(n_singlets, dtype=np.float64),
            singl_eng=np.zeros(n_singlets, dtype=np.float64),
        )

    def reset(self):
        """Reset all variables to zero."""
        self.quantities.reset()

    def get_current_energy(self) -> float:
        """Return the current energy of the attached calculator object."""
        return self.calc.results["energy"]

    def observe_step(self, mc_step: MCStep):
        """Update all SGC parameters."""
        # Reference to avoid self. lookup multiple times
        quantities = self.quantities
        E = self.get_current_energy()

        quantities.counter += 1
        quantities.energy += E
        quantities.energy_sq += E * E

        if self.observe_singlets:
            # Only perform the singlet observations if requested.
            new_singlets = self.calc.get_singlets()
            quantities.singlets += new_singlets
            quantities.singlets_sq += new_singlets * new_singlets
            quantities.singl_eng += new_singlets * E

    @property
    def energy(self):
        return self.quantities.energy

    @property
    def energy_sq(self):
        return self.quantities.energy_sq

    @property
    def singlets(self):
        return self.quantities.singlets

    @property
    def singl_eng(self):
        return self.quantities.singl_eng

    @property
    def counter(self):
        return self.quantities.counter

    def interval_ok(self, interval):
        return interval == 1



from typing import Optional, TextIO, Iterator
import logging
import sys
from contextlib import contextmanager

__all__ = ("log_stream", "log_stream_context", "get_root_clease_logger")


def get_root_clease_logger() -> logging.Logger:
    """Get the root clease logger."""
    return logging.getLogger("clease")


def log_stream(
    level: Optional[int] = None, stream: TextIO = sys.stdout, fmt: Optional[str] = None
) -> logging.StreamHandler:
    """Helper function to enable CLEASE logging to a stream. Default stream is stdout.
    This function permanently adjusts the global CLEASE logger. Use ``log_stream_context``
    to temporarily adjust the clease logger.
    Returns the added stream handler, so that it can be adjusted as needed.

    Following example will cause all logging statements of "INFO" or higher to be
    printed to stdout indefinitely.
    >>> log_stream(level=logging.INFO)

    Parameters:

    level: int or ``None``
        Sets the level of CLEASE logger. If ``None``, the level will not be adjusted.

    stream:
        Stream to add logging to, e.g. sys.stdout or an opened file.

    fmt: str
        Custom format message. See the python ``logging`` module for more details.
        Default format is '%(name)s - %(levelname)s - %(message)s'.
    """

    # Get the root CLEASE logger
    root_logger = get_root_clease_logger()

    handler = _make_stream_handler(level, stream, fmt=fmt)

    if level is not None:
        # Update the root log level
        root_logger.setLevel(level)
    root_logger.addHandler(handler)
    return handler


@contextmanager
def log_stream_context(
    level: Optional[int] = None, stream: TextIO = sys.stdout, fmt: Optional[str] = None
) -> Iterator[logging.StreamHandler]:
    """Context which temporarily adds a stream handler to the root CLEASE logger.
    Yields the added stream handler, so that it can be adjusted as needed.

    Example usage:
    with log_stream_context(level=logging.INFO):
        # CLEASE logs will also be printed to stdout
    # the stdout handler is removed, and logs will no longer be printed here.

    Parameters:

    level: int or ``None``
        Sets the level of CLEASE logger. If ``None``, the level will not be adjusted.

    stream:
        Stream to add logging to, e.g. sys.stdout or an opened file.

    fmt: str
        Custom format message. See the python ``logging`` module for more details.
        Default format is '%(name)s - %(levelname)s - %(message)s'.
    """

    root_logger = get_root_clease_logger()
    original_log_level = root_logger.level  # Current log level

    # Creates the handler, and adds it to the root logger
    handler = log_stream(level=level, stream=stream, fmt=fmt)

    try:
        # We yield the handler, so the user can make mutations
        # to the handler directly if they so choose, since it's attached directly
        # into the root logger.
        yield handler
    finally:
        # Remove the handler, and reset the log level (if needed)
        root_logger.removeHandler(handler)
        if level is not None:
            # We adjusted the root logger, reset it
            root_logger.setLevel(original_log_level)


def _make_stream_handler(
    level: Optional[int], stream: TextIO, fmt: Optional[str] = None
) -> logging.StreamHandler:
    """Helper function to create a stream handler with a specified level and stream.
    If the level is None, use the effective level of the CLEASE root logger.
    """
    # Find the effective level to give the handler
    eff_level = _get_effective_level(level)

    # Create the stream handler, and set the level and formatting.
    handler = logging.StreamHandler(stream)
    handler.setLevel(eff_level)
    # Set some formatting
    if fmt is None:
        fmt = "%(name)s - %(levelname)s - %(message)s"
    formatter = logging.Formatter(fmt)
    handler.setFormatter(formatter)
    return handler


def _get_effective_level(level: Optional[int]) -> int:
    """Helper function to get the effective level - if the level is None,
    we use the current level of the CLEASEroot logger, otherwise use the provided level
    """
    # If level is None, use the current level of the root logger
    if level is None:
        return get_root_clease_logger().level
    return level



from typing import Dict, Optional, Iterable, Set, List, Union, Sequence
from ase import Atoms
from clease.datastructures import SystemChange
from clease.settings import ClusterExpansionSettings
from .clease import Clease


class CleaseVolDep(Clease):
    """
    Calculator that can be used together with volume dependent ECIs

    :param settings: ClusterExpansionSettings object used to construct the
        cluster expansion
    :param eci: Dictionary with ECI. The keys in the dictionary should have the
        following form: c<size>_<diameter>_<id>_<decoration_number>_V<power>.
        Example: if the contribution to the total energy from the
        nearest neighbour pair cluster has the mathematical form

        (A + B*V + C*V^2)*CF,

        where CF is the nearest neighbour pair correlation function, the
        following coefficients should be in the ECI dict (the id is 0 and
        decoration number 00 in this example. They could be different for other
        correlation functions)

        {
            'c2_d0000_0_00_V0': A,
            'c2_d0000_0_00_V1': B,
            'c2_d0000_0_00_V2': C
        }

        The total energy is given by a sum of terms of the form depicted above.
        Example: If the total energy depends on the one-body, nearest neighbour
        pair cluster and shorted three body cluster with correlation functions
        CF1, CF2 and CF3, respectively. The expression for the total energy is

        E = (A1 + B1*V + C1*V^2)*CF1 + (A2 + B2*V + C2*V^2)*CF2 +
            (A3 + B3*V + C3*V^2)*CF3

        the ECI dictionary required to represent this expression is

        {
            'c1_0_V0': A1,
            'c1_0_V1': B1,
            'c1_0_V3': C1,
            'c2_d0000_0_00_V0': A2,
            'c2_d0000_0_00_V1': B2,
            'c2_d0000_0_00_V2': C2,
            'c3_d0000_0_000_V0': A3,
            'c3_d0000_0_000_V1': B3,
            'c3_d0000_0_000_V2': C3,
        }

        again the id has been set to 0 and the decoration numbers are only 00
        (and 000), which will vary for different types of correlation functions.
    :param vol_coeff: Expansion coefficients used to cluster expand the volume.
        The volume coefficients should not have the V<power> tag at the end.
        Example: If the volume can be described by the following mathematical
        form

        V = A + B*CF1 + C*CF2,

        where CF1 is the correlation function of one-body clusters and CF2 is
        the correlation function of the nearest neighbour pair interacation,
        the passed dictionary should be

        {
            'c0': A,
            'c1_0': B,
            'c2_d0000_0_00': C
        }
    :param init_cf: Correlation functions of the initial atoms object. They can
        for instance be obtained via `clease.CorrFunction`. If not passed,
        they are calculated from scratch when an atoms object is attached.
    """

    def __init__(
        self,
        settings: ClusterExpansionSettings,
        eci: Dict[str, float],
        vol_coeff: Dict[str, float],
        init_cf: Optional[Dict[str, float]] = None,
    ):

        if not eci_format_ok(eci.keys()):
            raise ValueError(f"Invalid format of ECI names. Got\n{eci.keys()}")

        if not vol_coeff_format_ok(vol_coeff.keys()):
            raise ValueError(
                "Invalid format of volume coefficient names. " f"Got\n{vol_coeff.keys()}"
            )

        cf_to_track = unique_eci_names_no_vol(eci.keys())
        cf_to_track = cf_to_track.union(set(vol_coeff.keys()))

        # Create a set of empty ECI and pass to the parent class.
        # This makes sure that all nessecary correlation functions
        # are tracked.
        eci_track = {k: 0.0 for k in cf_to_track}

        # Transfer the volume independent part to the parent calculator
        for k, v in eci.items():
            if k.endswith("V0"):
                key = k.rpartition("_V0")[0]
                eci_track[key] = v

        Clease.__init__(self, settings, eci_track, init_cf=init_cf)
        self.eci_with_vol = eci
        self.vol_coeff = vol_coeff
        self.max_power = max(int(k[-1]) for k in self.eci_with_vol.keys())

    def get_volume(self, cf: Optional[Dict[str, float]] = None) -> float:
        """
        Returns the volume per atom

        :param cf: Correlation functions. If not given, the correlation
            functions are updated to match the current state of the attached
            atoms object
        """
        if cf is None:
            self.update_cf(None)
            cf = self.get_cf()
        vol = sum(self.vol_coeff[k] * cf[k] for k in self.vol_coeff)
        self.results["volume_per_atom"] = vol
        self.results["volume"] = vol * len(self.atoms)
        return vol

    def get_pressure(self, cf: Optional[Dict[str, float]] = None) -> float:
        """
        Return the pressure

        :param cf: Correlation functions. If not given, the correlation
            functions are updated to match the current state of the attached
            atoms object
        """
        if cf is None:
            self.update_cf(None)
            cf = self.get_cf()

        vol = self.get_volume(cf)
        P = sum(
            p * self.eci_with_vol.get(k + f"_V{p}", 0.0) * cf[k] * vol ** (p - 1)
            for k in cf.keys()
            for p in range(1, self.max_power + 1)
        )
        self.results["pressure"] = P
        return P

    def get_bulk_modulus(self, cf: Optional[Dict[str, float]] = None) -> float:
        """
        Return the bulk modulus of the current atoms object

        :param cf: Correlation functions. If not given, the correlation
            functions are updated to match the current state of the attached
            atoms object
        """
        if cf is None:
            self.update_cf(None)
            cf = self.get_cf()

        vol = self.get_volume(cf)
        B = vol * self._d2EdV2(cf, vol)
        self.results["bulk_mod"] = B
        return B

    def _d2EdV2(self, cf: Dict[str, float], vol: float) -> float:
        """
        Return the double derivative of the energy with respect
        to the volume

        :param cf: Correlation functions
        :param vol: Volume
        """
        return sum(
            p * (p - 1) * self.eci_with_vol.get(k + f"_V{p}", 0.0) * cf[k] * vol ** (p - 2)
            for k in cf.keys()
            for p in range(2, self.max_power + 1)
        )

    def _d3EdV3(self, cf: Dict[str, float], vol: float) -> float:
        """
        Return the third derivative of the energy with respect to
        the volume

        :param cf: Correlation functions
        :param vol: Volume
        """
        return sum(
            p
            * (p - 1)
            * (p - 2)
            * self.eci_with_vol.get(k + f"_V{p}", 0.0)
            * cf[k]
            * vol ** (p - 3)
            for k in cf.keys()
            for p in range(3, self.max_power + 1)
        )

    def get_dBdP(self, cf: Dict[str, float] = None) -> float:
        """
        Return the pressure derivative of the bulk modulus of the
        current structure.

        :param cf: Correlation functions. If not given, the correlation
            functions are updated to match the current state of the attached
            atoms object
        """
        if cf is None:
            self.update_cf(None)
            cf = self.get_cf()
        vol = self.get_volume(cf)
        return -1.0 - vol * self._d3EdV3(cf, vol) / self._d2EdV2(cf, vol)

    def get_energy_given_change(self, system_changes: Sequence[SystemChange], keep_changes=False):
        """
        Update correlation functions given the change
        """
        with self.with_system_changes(system_changes) as keeper:
            keeper.keep_changes = keep_changes
            cf = self.get_cf()
            vol = self.get_volume(cf)
            energy = sum(
                self.eci_with_vol.get(k + f"_V{p}", 0.0) * cf[k] * vol**p
                for k in cf.keys()
                for p in range(self.max_power + 1)
            )
            self.energy = energy * len(self.atoms)
            self.results["energy"] = self.energy
            return self.energy

    def calculate(
        self,
        atoms: Atoms = None,
        properties: List[str] = None,
        system_changes: Union[Sequence[SystemChange], None] = None,
    ) -> float:
        """Calculate the energy of the passed Atoms object.

        Parameters:

        :param atoms: ASE Atoms object, or None. If None is passed,
            the internal one is used, if the calculator has been initialized.

        :param system_changes:
            System changes. See doc-string of
            `clease.montecarlo.observers.MCObserver`
        """
        super().calculate(atoms, properties, system_changes)
        self.update_cf()
        return self.get_energy_given_change([], keep_changes=True)

    def _on_eci_changed(self):
        """
        Callback triggered by the parent class when the ECIs are changed.
        """
        for k, v in self.eci.items():
            vol_dep_key = k + "_V0"
            self.eci_with_vol[vol_dep_key] = v


def unique_eci_names_no_vol(eci_names: Iterable[str]) -> Set[str]:
    """
    Return a set with the unique ECI names without any volume tag
    """
    eci_no_vol = set(k.rpartition("_V")[0] for k in eci_names)
    return eci_no_vol


def eci_format_ok(eci_names: Iterable[str]) -> bool:
    """
    Check that all ECIs are formatted correctly.
    """
    for name in eci_names:
        valid = name.startswith("c") and "_V" in name
        if not valid:
            return False
    return True


def vol_coeff_format_ok(vol_coeff_names: Iterable[str]) -> bool:
    """
    Check that the volume coefficients are valid
    """
    for name in vol_coeff_names:
        valid = name.startswith("c") and "_V" not in name
        if not valid:
            return False
    return True



"""Expose the C++ Atoms object to Python"""
# distutils: language = c++
#import copy
from typing import List, Sequence
from atoms cimport Atoms
from libcpp.vector cimport vector
# 4-vector, as per the C++ definiton
from four_vector cimport FourVector as cppFourVector
# The CLEASE Python definiton of the 4-vector
from clease.datastructures.four_vector import FourVector
from clease.datastructures.system_changes import SystemChange
import ase

cdef class CppAtoms:
    cdef Atoms *thisptr

    def __cinit__(self, atoms: ase.Atoms, four_vectors: Sequence[FourVector]):
        # Extract the atomic numbers directly from the Atoms object,
        # and make a list-copy.
        #numbers = list(atoms.get_atomic_numbers())
        self.thisptr = new Atoms(atoms, four_vectors)

    def __dealloc__(self):
        del self.thisptr

    def get_numbers(self) -> List[int]:
        return list(self.thisptr.get_numbers())

    def get_symbols(self) -> List[str]:
        return list(self.thisptr.get_symbols())

    def get_four_vectors(self) -> List[FourVector]:
        cdef vector[cppFourVector] vectors
        vectors = self.thisptr.get_four_vectors()

        L = vectors.size()
        return [_to_clease_fourvec(vectors[i]) for i in range(L)]

    @property
    def Ns(self) -> int:
        return self.thisptr.Ns

    @property
    def Nx(self) -> int:
        return self.thisptr.Nx

    @property
    def Ny(self) -> int:
        return self.thisptr.Ny

    @property
    def Nz(self) -> int:
        return self.thisptr.Nz

    def get_1d_index(self, vec: FourVector) -> int:
        cdef cppFourVector cpp_vec
        cdef int idx
        cpp_vec = _to_cpp_fourvec(vec)
        idx = self.thisptr.get_1d_index(cpp_vec)
        return idx

    def apply_change(self, change: SystemChange) -> None:
        self.thisptr.apply_change(change)

    def undo_change(self, change: SystemChange) -> None:
        self.thisptr.undo_change(change)

cdef _to_clease_fourvec(cppFourVector vec):
    """Convert a c++ FourVector into a CLEASE Python FourVector object."""
    return FourVector(vec.ix, vec.iy, vec.iz, vec.sublattice)

cdef cppFourVector _to_cpp_fourvec(vec: FourVector):
    cdef cppFourVector fourvec
    fourvec = cppFourVector()
    fourvec.ix = vec.ix
    fourvec.iy = vec.iy
    fourvec.iz = vec.iz
    fourvec.sublattice = vec.sublattice
    return fourvec



from typing import Sequence, Dict, Any
import numpy as np
from ase import Atoms
from ase.units import kB
from clease.calculator import Clease
from clease.settings import ClusterExpansionSettings
from .observers import SGCObserver
from .montecarlo import Montecarlo
from .trial_move_generator import TrialMoveGenerator, RandomFlip


class InvalidChemicalPotentialError(Exception):
    pass


class SGCMonteCarlo(Montecarlo):
    """
    Class for running Monte Carlo in the Semi-Grand Canonical Ensebmle
    (i.e., fixed number of atoms, but varying composition)

    See the docstring of :class:`clease.montecarlo.Montecarlo`

    :param atoms: Atoms object (with CLEASE calculator attached!)
    :param temp: Temperature in kelvin
    :param symbols: Possible symbols to be used in swaps
    :param generator: Generator that produces trial moves
    """

    def __init__(
        self,
        atoms: Atoms,
        temp: float,
        symbols: Sequence[str] = (),
        generator: TrialMoveGenerator = None,
        observe_singlets: bool = False,
    ):
        if not isinstance(atoms, Atoms):
            raise ValueError(f"atoms must be an Atoms object, got {atoms!r}")
        if not isinstance(atoms.calc, Clease):
            raise ValueError(
                f"Atoms must have a Clease calculator object attached, got {atoms.calc!r}."
            )

        if generator is None:
            if len(symbols) <= 1:
                raise ValueError("At least 2 symbols have to be specified")
            # Only select indices which are not considered background.
            non_bkg = atoms.calc.settings.non_background_indices
            generator = RandomFlip(symbols, atoms, indices=non_bkg)

        self.averager = SGCObserver(atoms.calc, observe_singlets=observe_singlets)
        super().__init__(atoms, temp, generator=generator)

        self.symbols = symbols

        self.chem_pots = []
        self.chem_pot_names = []
        self.has_attached_avg = False
        self.name = "SGCMonteCarlo"
        self._chemical_potential = None
        self.chem_pot_in_eci = False

        has_attached_obs = False
        for obs in self.iter_observers():
            if obs.name == "SGCObserver":
                has_attached_obs = True
                self.averager = obs
                break
        if not has_attached_obs:
            self.attach(self.averager)

    @property
    def observe_singlets(self) -> bool:
        return self.averager.observe_singlets

    def _check_symbols(self):
        """
        Override because there are no restriction on the symbols here
        """

    def reset(self):
        """
        Reset the simulation object
        """
        super().reset()
        self.averager.reset()

    @property
    def calc(self) -> Clease:
        return self.atoms.calc

    @property
    def settings(self) -> ClusterExpansionSettings:
        return self.calc.settings

    @property
    def chemical_potential(self):
        return self._chemical_potential

    @chemical_potential.setter
    def chemical_potential(self, chem_pot: Dict[str, float]):
        eci = self.calc.eci
        if any(key not in eci for key in chem_pot):
            msg = "A chemical potential not being trackted is added. Make "
            msg += "sure that all the following keys are in the ECIs before "
            msg += "they are passed to the calculator: "
            msg += f"{list(chem_pot.keys())}\n"
            msg += "(Add them with a zero ECI value if they are not supposed "
            msg += "to be included.)"
            raise InvalidChemicalPotentialError(msg)

        self._chemical_potential = chem_pot
        if self.chem_pot_in_eci:
            self._reset_eci_to_original(self.calc.eci)
        self._include_chemical_potential_in_eci(chem_pot, self.calc.eci)

    def _include_chemical_potential_in_eci(self, chem_pot: Dict[str, float], eci: Dict[str, float]):
        """
        Including the chemical potentials in the ECIs

        Parameters:


        chem_pot: dict
            Chemical potentials
        eci: dict
            Original ECIs
        """
        self.chem_pots = []
        self.chem_pot_names = []
        keys = list(chem_pot.keys())
        keys.sort()
        for key in keys:
            self.chem_pots.append(chem_pot[key])
            self.chem_pot_names.append(key)
            current_eci = eci.get(key, 0.0)
            eci[key] = current_eci - chem_pot[key]
        calc = self.calc
        calc.update_eci(eci)
        self.chem_pot_in_eci = True
        self.current_energy = calc.calculate(None, None, None)
        return eci

    def _reset_eci_to_original(self, eci_with_chem_pot: Dict[str, float]):
        """
        Resets the ECIs to their original value

        :parma dict eci_with_chem_pot: ECIs with chemical potential included
        """
        for name, val in zip(self.chem_pot_names, self.chem_pots):
            eci_with_chem_pot[name] += val
        calc = self.calc
        calc.update_eci(eci_with_chem_pot)
        self.chem_pot_in_eci = False
        self.current_energy = calc.calculate(None, None, None)
        return eci_with_chem_pot

    def reset_eci(self):
        """Return the ECIs."""
        if self.chem_pot_in_eci:
            self._reset_eci_to_original(self.calc.eci)

    def run(self, steps: int = 10, call_observers: bool = True, chem_pot: Dict[str, float] = None):
        """
        Run Monte Carlo simulation.
        See :py:meth:`~clease.montecarlo.montecarlo.Montecarlo.run`

        Parameters:

        chem_pot: dict
            Chemical potentials. The keys should correspond to one of the
            singlet terms. A typical form of this is {"c1_0":-1.0,c1_1_1.0}
        """

        if chem_pot is None and self.chemical_potential is None:
            ex_chem_pot = {"c1_1": -0.1, "c1_2": 0.05}
            raise ValueError(
                f"No chemical potentials given. Has to be a dictionary of the form {ex_chem_pot}"
            )

        if chem_pot is not None:
            self.chemical_potential = chem_pot
        self.reset()

        super().run(steps=steps, call_observers=call_observers)

    def singlet2composition(self, avg_singlets: Dict[str, float]):
        """Convert singlets to composition."""
        bf = self.settings.basis_functions
        matrix = np.zeros((len(self.symbols), len(self.symbols)))

        index = {s: i for i, s in enumerate(self.symbols)}
        for i, b in enumerate(bf):
            for s, col in index.items():
                matrix[i, col] = b[s]

        matrix[-1, :] = 1.0
        rhs = np.zeros(len(self.symbols))
        rhs[:-1] = avg_singlets
        rhs[-1] = 1.0
        x = np.linalg.solve(matrix, rhs)

        res = {}
        for s, i in index.items():
            name = f"{s}_conc"
            res[name] = x[i]
        return res

    def reset_averagers(self) -> None:
        """Reset the energy averagers, including the internal SGC Observer"""
        super().reset_averagers()
        # Also remember to reset the internal SGC averager
        self.averager.reset()

    def get_thermodynamic_quantities(self, reset_eci: bool = False) -> Dict[str, Any]:
        """Compute thermodynamic quantities.

        Parameters:

        reset_eci: bool
            If True, the chemical potential will be removed from the ECIs.
        """
        # Note - in order to correctly measure averagers from the SGC observer,
        # we need some information from the SGC MC object. So we directly extract the averages
        # from that observer here.
        avg_obs = self.averager  # Alias
        N = self.averager.counter
        averages = {}
        averages["energy"] = avg_obs.energy.mean
        averages["sgc_energy"] = avg_obs.energy.mean
        averages["sgc_heat_capacity"] = avg_obs.energy_sq.mean - avg_obs.energy.mean**2

        averages["sgc_heat_capacity"] /= kB * self.temperature**2

        averages["temperature"] = self.temperature
        averages["n_mc_steps"] = self.averager.counter
        averages["accept_rate"] = self.current_accept_rate

        # Singlets are more expensive to measure than the other quantities,
        # so only measure them upon request.
        if self.observe_singlets:
            # Add singlets and chemical potential to the dictionary
            # pylint: disable=consider-using-enumerate
            singlets = avg_obs.singlets / N
            singlets_sq = avg_obs.quantities["singlets_sq"] / N

            averages["singlet_energy"] = avg_obs.energy.mean
            natoms = len(self.atoms)
            for i, chem_pot in enumerate(self.chem_pots):
                averages["singlet_energy"] += chem_pot * singlets[i] * natoms
            for i in range(len(singlets)):
                name = f"singlet_{self.chem_pot_names[i]}"
                averages[name] = singlets[i]

                name = f"var_singlet_{self.chem_pot_names[i]}"
                averages[name] = singlets_sq[i] - singlets[i] ** 2

            # Measure concentration from the singlets
            try:
                avg_conc = self.singlet2composition(singlets)
                averages.update(avg_conc)
            # pylint: disable=broad-except
            except Exception as exc:
                print("Could not find average singlets!")
                print(exc)

        # Always measure the chemical potentials.
        for chem_pot_name, chem_pot in zip(self.chem_pot_names, self.chem_pots):
            key = f"mu_{chem_pot_name}"
            averages[key] = chem_pot

        averages.update(self.meta_info)

        # Add information from observers
        averages.update(self._get_obs_averages())

        if reset_eci:
            self._reset_eci_to_original(self.atoms.calc.eci)
        return averages



import logging
import numpy as np
import ase
from clease.datastructures import MCStep
from clease.calculator import CleaseCacheCalculator
from .mc_observer import MCObserver

logger = logging.getLogger(__name__)


class LowestEnergyStructure(MCObserver):
    """Track the lowest energy state visited during an MC run.

    atoms: Atoms object
        Atoms object used in Monte Carlo

    track_cf: bool
        Whether to keep a copy of the correlation functions for the
        emin structure. If enabled, this will be stored in the ``lowest_energy_cf``
        variable. Defaults to False.

    verbose: bool
        If `True`, progress messages will be printed
    """

    name = "LowestEnergyStructure"

    def __init__(self, atoms: ase.Atoms, track_cf: bool = False, verbose: bool = False):
        super().__init__()
        self.atoms = atoms
        self.track_cf = track_cf
        self.verbose = verbose
        # Silence pylint
        self.lowest_energy_cf = None
        self.reset()

    def reset(self) -> None:
        self.lowest_energy_cf = None
        self.emin_atoms: ase.Atoms = self.atoms.copy()
        # Keep a reference directly to the calculator cache
        self._calc_cache = CleaseCacheCalculator()
        self.emin_atoms.calc = self._calc_cache
        self.lowest_energy = np.inf

    @property
    def lowest_energy(self) -> float:
        return self.emin_results["energy"]

    @lowest_energy.setter
    def lowest_energy(self, value: float) -> None:
        """The currently best observed energy"""
        self.emin_results["energy"] = value
        # Keep a copy without needing to look up the calculator
        # all the time, for quicker lookup.
        self._lowest_energy_cache = value

    @property
    def emin_results(self) -> dict:
        """The results dictionary of the lowest energy atoms"""
        return self._calc_cache.results

    @property
    def calc(self):
        return self.atoms.calc

    @property
    def energy(self):
        """The energy of the current atoms object (not the emin energy)"""
        return self.calc.results["energy"]

    def observe_step(self, mc_step: MCStep) -> None:
        """
        Check if the current state has lower energy and store the current
        state if it has a lower energy than the previous state.

        mc_step: MCStep
             Instance of MCStep with information on the latest step.
        """
        # We do the comparison even if the move was rejected,
        # as no energy has been recorded for the first step, yet.
        # The internal energy cache is np.inf when nothing has been recorded.
        if mc_step.energy < self._lowest_energy_cache:
            self._update_emin(mc_step)
            dE = mc_step.energy - self.lowest_energy
            msg = "Found new low energy structure. New energy: %.3f eV. Change: %.3f eV"
            logger.info(msg, self.lowest_energy, dE)
            if self.verbose:
                print(msg % (self.lowest_energy, dE))

    def _update_emin(self, mc_step: MCStep) -> None:
        if self.track_cf:
            self.lowest_energy_cf = self.calc.get_cf()
        self.lowest_energy = mc_step.energy

        # Avoid copying the atoms object multiple times. We can only ever have
        # changed the symbols, so copy those over from the atoms object.
        self.emin_atoms.numbers = self.atoms.numbers



"""Definition of ClusterExpansionSettings Class.

This module defines the base-class for storing the settings for performing
Cluster Expansion in different conditions.
"""
from __future__ import annotations
import logging
from copy import deepcopy
from typing import List, Dict, Optional, Union, Sequence, Set, Any

import numpy as np
from ase import Atoms
from ase.db import connect
from ase.db.core import Database

from clease.version import __version__
from clease.jsonio import jsonable
from clease.tools import get_size_from_cf_name, wrap_and_sort_by_position
from clease.cluster import ClusterManager, ClusterList, Cluster
from clease.basis_function import Polynomial, Trigonometric, BinaryLinear, BasisFunction
from clease.datastructures import TransMatrix
from .concentration import Concentration
from .template_filters import ValidConcentrationFilter
from .template_atoms import TemplateAtoms
from .atoms_manager import AtomsManager


__all__ = ("ClusterExpansionSettings",)

logger = logging.getLogger(__name__)


class PrimitiveCellNotFound(Exception):
    """Exception which is raised if the primitive isn't found in the database."""


@jsonable("ce_settings")
class ClusterExpansionSettings:
    """Base class for all Cluster Expansion settings.

    Args:
        prim (Atoms): The primitive atoms object.

        concentration (Union[Concentration, dict]): Concentration object or
            dictionary specifying the basis elements and
            concentration range of constituting species.

        size (List[int] | None, optional): Size of the supercell
            (e.g., [2, 2, 2] for 2x2x2 cell).
            ``supercell_factor`` is ignored if both ``size`` and ``supercell_factor``
            are specified. Defaults to None.

        supercell_factor (int, optional): Maximum multipilicity factor for
            limiting the size of supercell created from the primitive cell.
            ``supercell_factor`` is ignored if
            both `size` and `supercell_factor` are specified. Defaults to 27.

        db_name (str, optional): Name of the database file. Defaults to ``'clease.db'``.

        max_cluster_dia (Sequence[float], optional): A list of int or float containing the
            maximum diameter of clusters (in Å). Defaults to ``(5., 5., 5.)``, i.e.
            a 5 Å cutoff for 2-, 3-, and 4-body clusters.

        include_background_atoms (bool, optional): Whether background elements are to
            be included. An element is considered to be a background element,
            if there is only 1 possible species which be ever be placed in a given basis.
            Defaults to False.

        basis_func_type (str, optional): Type of basis function to use. Defaults to 'polynomial'.
    """

    # pylint: disable=too-many-instance-attributes, too-many-public-methods

    # Keys which are important for saving/loading
    ARG_KEYS = ("prim_cell", "concentration")
    KWARG_KEYS = (
        "size",
        "supercell_factor",
        "db_name",
        "max_cluster_dia",
        "include_background_atoms",
        "basis_func_type",
    )

    # Other keys we want to input back into the loaded object, but after initialization
    OTHER_KEYS = (
        "skew_threshold",
        # kwargs is a bookkeeping variable, for compatibility.
        # Just contains information about how the factory functions were called.
        "kwargs",
    )

    def __init__(
        self,
        prim: Atoms,
        concentration: Union[Concentration, dict],
        size: Optional[List[int]] = None,
        supercell_factor: Optional[int] = 27,
        db_name: str = "clease.db",
        max_cluster_dia: Sequence[float] = (5.0, 5.0, 5.0),
        include_background_atoms: bool = False,
        basis_func_type="polynomial",
    ) -> None:
        self._include_background_atoms = include_background_atoms
        self._cluster_mng = None
        self._trans_matrix = None
        self._cluster_list = None
        self._basis_func_type = None
        self._prim_cell = None
        self.concentration = _get_concentration(concentration)

        self.basis_elements = deepcopy(self.concentration.basis_elements)
        self._check_first_elements()

        self.db_name = db_name
        self._set_prim_cell(prim)

        prim_mng = AtomsManager(self.prim_cell)
        prim_ind_by_basis = prim_mng.index_by_symbol([x[0] for x in self.basis_elements])
        conc_filter = ValidConcentrationFilter(concentration, prim_ind_by_basis)

        self.atoms_mng = AtomsManager(None)

        # Construct the template atoms, this is a protected property
        # Access and changes to size and/or supercell factor are redirected to
        # this instance.
        self._template_atoms = TemplateAtoms(
            self.prim_cell,
            supercell_factor=supercell_factor,
            size=size,
            skew_threshold=40,
            filters=[conc_filter],
        )

        self.set_active_template()

        self.max_cluster_dia = max_cluster_dia

        self.basis_func_type = basis_func_type

        if len(self.basis_elements) != self.num_basis:
            raise ValueError("list of elements is needed for each basis")

        # For storing the settings from the CLEASE factories.
        # Just for bookkeeping
        self.kwargs = {}

    @property
    def atoms(self) -> Atoms:
        """The currently active template."""
        return self.atoms_mng.atoms

    @property
    def prim_cell(self) -> Atoms:
        """The primitive atoms object of the model."""
        return self._prim_cell

    def _set_prim_cell(self, value: Atoms) -> None:
        """Set the primitive cell, ensure it gets properly tagged and stored.
        This should not be changed after initialization."""
        self._prim_cell = value.copy()
        self._prim_cell.wrap()
        self._order_and_tag_prim_cell()
        self._ensure_primitive_exists()

    @property
    def max_cluster_dia(self) -> np.ndarray:
        """The maximum cluster diameter, expressed in a NumPy
        array starting from 2-body clusters at index 0.
        Diameters are given in units of Ångstrom.
        """
        return self._max_cluster_dia

    @max_cluster_dia.setter
    def max_cluster_dia(self, value) -> None:
        # Ensure it's NumPy array, and make a copy.
        self._max_cluster_dia = np.array(value, dtype=float)
        self.clear_cache()

    @property
    def db_name(self) -> str:
        """Name of the underlaying data base."""
        return self._db_name

    @db_name.setter
    def db_name(self, value: str) -> str:
        """Changing the DB name, needs to ensure the primitive cell exists in
        the new database as well."""
        self._db_name = value
        if self.prim_cell is not None:
            # Ensure the primitive cell is stored in the new database.
            # None should only happen during initialization, where we let the
            # prim_cell setter store the primitive.
            self._ensure_primitive_exists()

    def connect(self, **kwargs) -> Database:
        """Return the ASE connection object to the internal database."""
        return connect(self.db_name, **kwargs)

    @property
    def template_atoms(self) -> TemplateAtoms:
        return self._template_atoms

    @property
    def max_cluster_size(self):
        return len(self.max_cluster_dia) + 1

    @property
    def all_elements(self) -> List[str]:
        return sorted([item for row in self.basis_elements for item in row])

    @property
    def num_elements(self) -> int:
        return len(self.all_elements)

    @property
    def unique_elements(self) -> List[str]:
        return sorted(list(set(deepcopy(self.all_elements))))

    @property
    def num_unique_elements(self) -> int:
        return len(self.unique_elements)

    @property
    def ref_index_trans_symm(self) -> List[int]:
        return [i[0] for i in self.index_by_sublattice]

    @property
    def skew_threshold(self):
        return self.template_atoms.skew_threshold

    @skew_threshold.setter
    def skew_threshold(self, threshold: int) -> None:
        """
        Maximum acceptable skew level (ratio of max and min diagonal of the
        Niggli reduced cell)
        """
        self.template_atoms.skew_threshold = threshold

    @property
    def background_indices(self) -> List[int]:
        """Get indices of the background atoms."""
        # check if any basis consists of only one element type
        basis = [i for i, b in enumerate(self.basis_elements) if len(b) == 1]

        bkg_indices = []
        for b_indx in basis:
            bkg_indices += self.index_by_basis[b_indx]
        return bkg_indices

    @property
    def non_background_indices(self) -> List[int]:
        """Indices of sites which are not background"""
        bkg = set(self.background_indices)

        all_indices = set(range(len(self.atoms)))
        # Remove all background indices
        return sorted(all_indices - bkg)

    @property
    def cluster_mng(self):
        if self._cluster_mng is None:
            kwargs = {}
            if not self.include_background_atoms:
                kwargs["background_syms"] = self.get_bg_syms()
            self._cluster_mng = ClusterManager(self.prim_cell, **kwargs)
        return self._cluster_mng

    @property
    def include_background_atoms(self) -> bool:
        return self._include_background_atoms

    @include_background_atoms.setter
    def include_background_atoms(self, value: bool) -> None:
        # pylint: disable=no-self-use
        msg = "The include_background_atoms setter has been removed in version 0.11.3.\n"
        msg += f"Please set 'include_background_atoms={value}' in the settings constructor, "
        msg += "instead."
        raise NotImplementedError(msg)

    @property
    def spin_dict(self) -> Dict[str, float]:
        return self.basis_func_type.spin_dict

    @property
    def basis_functions(self):
        return self.basis_func_type.basis_functions

    @property
    def ignore_background_atoms(self) -> bool:
        return not self.include_background_atoms

    @property
    def multiplicity_factor(self) -> Dict[str, float]:
        """Return the multiplicity factor of each cluster."""
        num_sites_in_group = [len(x) for x in self.index_by_sublattice]
        return self.cluster_list.multiplicity_factors(num_sites_in_group)

    @property
    def all_cf_names(self) -> List[str]:
        num_bf = len(self.basis_functions)
        return self.cluster_list.get_all_cf_names(num_bf)

    @property
    def num_cf(self) -> int:
        """Return the number of correlation functions."""
        return len(self.all_cf_names)

    @property
    def index_by_basis(self) -> List[List[int]]:
        first_symb_in_basis = [x[0] for x in self.basis_elements]
        return self.atoms_mng.index_by_symbol(first_symb_in_basis)

    @property
    def index_by_sublattice(self) -> List[List[int]]:
        return self.atoms_mng.index_by_tag()

    @property
    def num_basis(self) -> int:
        return len(self.basis_elements)

    @property
    def basis_func_type(self):
        return self._basis_func_type

    @property
    def size(self) -> Union[np.ndarray, None]:
        return self.template_atoms.size

    @size.setter
    def size(self, value):
        """Exposure of the template atoms size setter"""
        self.template_atoms.size = value

    @property
    def supercell_factor(self) -> Union[int, None]:
        return self.template_atoms.supercell_factor

    @supercell_factor.setter
    def supercell_factor(self, value) -> None:
        """Exposure of the template atoms supercell_factor setter"""
        self.template_atoms.supercell_factor = value

    def get_sublattice_site_ratios(self) -> np.ndarray:
        """Return the ratios of number of sites per (grouped) sublattice"""
        # Number of sites per sublattice
        sites_per_basis = np.array([len(basis) for basis in self.index_by_basis])
        num_sites = sites_per_basis.sum()
        return sites_per_basis / num_sites

    @property
    def num_active_sublattices(self) -> int:
        """Number of active sublattices"""
        return sum(self.get_active_sublattices())

    def get_active_sublattices(self) -> List[bool]:
        """List of booleans indicating if a (grouped) sublattice is active"""
        unique_no_bkg = self.unique_element_without_background()

        return [basis[0] in unique_no_bkg for basis in self.concentration.basis_elements]

    @property
    def ignored_species_and_conc(self) -> Dict[str, float]:
        """
        Return the ignored species and their concentrations normalised to the total number
        of atoms.
        """
        unique_no_bkg = self.unique_element_without_background()
        # Find the concentration within grouped sublattices
        orig_basis = self.concentration.basis_elements
        # Concentration of sites within each sublattice
        ratios = self.get_sublattice_site_ratios()
        assert len(orig_basis) == len(ratios)
        ignored = {}
        for ratio, basis in zip(ratios, orig_basis):
            elem = basis[0]
            if elem not in unique_no_bkg:
                if len(basis) != 1:
                    raise ValueError(
                        (
                            "Ignored sublattice contains multiple elements -"
                            "this does not make any sense"
                        )
                    )
                if elem not in ignored:
                    ignored[elem] = ratio
                else:
                    # This element is already on one of the ignored background here we
                    # accumulate the concentration
                    ignored[elem] += ratio
        return ignored

    @property
    def atomic_concentration_ratio(self) -> float:
        """
        Ratio between true concentration (normalised to atoms) and the internal concentration used.
        For example, if one of the two basis is fully occupied, and hence ignored internally, the
        internal concentration is half of the actual atomic concentration.
        """
        ratios = self.get_sublattice_site_ratios()
        # We only want to include active sublattices
        active_sublatt = self.get_active_sublattices()
        # Add up all ratios for the active sublattices
        return ratios[active_sublatt].sum()

    @basis_func_type.setter
    def basis_func_type(self, bf_type):
        """
        Type of basis function to use.
        It should be one of "polynomial", "trigonometric" or "binary_linear"
        """
        unique_element = self.unique_element_without_background()

        if isinstance(bf_type, BasisFunction):
            if bf_type.unique_elements != sorted(unique_element):
                raise ValueError(
                    "Unique elements in BasisFunction instance "
                    "is different from the one in settings",
                    bf_type.unique_elements,
                    sorted(unique_element),
                )
            self._basis_func_type = bf_type
        elif isinstance(bf_type, str):
            if bf_type.lower() == "polynomial":
                self._basis_func_type = Polynomial(unique_element)
            elif bf_type.lower() == "trigonometric":
                self._basis_func_type = Trigonometric(unique_element)
            elif bf_type.lower() == "binary_linear":
                self._basis_func_type = BinaryLinear(unique_element)
            else:
                msg = f"basis function type {bf_type} is not supported."
                raise ValueError(msg)
        else:
            raise ValueError("basis_function has to be an instance of BasisFunction or a string")

    def get_bg_syms(self) -> Set[str]:
        """
        Return the symbols in the basis where there is only one element
        """
        return set(x[0] for x in self.basis_elements if len(x) == 1)

    def unique_element_without_background(self):
        """Remove background elements."""
        if self.include_background_atoms:
            bg_sym = set()
        else:
            bg_sym = self.get_bg_syms()

            # Remove bg_syms that are also present in basis with more than one
            # element
            for elems in self.basis_elements:
                if len(elems) == 1:
                    continue
                to_be_removed = set()
                for s in bg_sym:
                    if s in elems:
                        to_be_removed.add(s)

                bg_sym -= to_be_removed

        unique_elem = set()
        for x in self.basis_elements:
            unique_elem.update(x)
        return list(unique_elem - bg_sym)

    def prepare_new_active_template(self, template):
        """Prepare necessary data structures when setting new template."""
        logger.debug("Preparing new template in settings")
        self.atoms_mng.atoms = template
        self.clear_cache()

    def set_active_template(self, atoms=None):
        """Set a new template atoms object."""
        if atoms is not None:
            template = self.template_atoms.get_template_matching_atoms(atoms=atoms)
        else:
            template = self.template_atoms.weighted_random_template()

        template = wrap_and_sort_by_position(template)

        if atoms is not None:
            # Check that the positions of the generated template
            # matches the ones in the passed object
            atoms = wrap_and_sort_by_position(atoms)
            if not np.allclose(template.get_positions(), atoms.get_positions()):
                raise ValueError(
                    f"Inconsistent positions. Passed object\n"
                    f"{atoms.get_positions()}\nGenerated template"
                    f"\n{template.get_positions()}"
                )
        self.prepare_new_active_template(template)

    def _order_and_tag_prim_cell(self):
        """
        Add a tag to all the atoms in the unit cell to track the sublattice.
        Tags are added such that that the lowest tags corresponds to "active"
        sites, while the highest tags corresponds to background sites. An
        example is a system having three sublattices that can be occupied by
        more than one species, and two sublattices that can be occupied by
        only one species. The tags 0, 1, 2 will then be assigned to the three
        sublattices that can be occupied by more than one species, and the two
        remaining lattices will get the tag 3, 4.

        Re-orders the primitive cell in accordance to the order of the tags.
        """
        bg_sym = self.get_bg_syms()
        tag = 0

        # Tag non-background elements first
        for atom in self.prim_cell:
            if atom.symbol not in bg_sym:
                atom.tag = tag
                tag += 1

        # Tag background elements
        for atom in self.prim_cell:
            if atom.symbol in bg_sym:
                atom.tag = tag
                tag += 1

        # Rearange primitive cell in order of the tags.
        sorted_indx = np.argsort([atom.tag for atom in self.prim_cell])
        # Set the primitive cell directly, without calling the setter.
        self._prim_cell = self.prim_cell[sorted_indx]

    def _ensure_primitive_exists(self) -> None:
        """Ensure that the primitive cell exists in the DB.
        Write it if it is missing."""
        self.get_prim_cell_id(write_if_missing=True)

    def get_prim_cell_id(self, write_if_missing=False) -> int:
        """Retrieve the ID of the primitive cell in the database.
        Raises a PrimitiveCellNotFound error if it is not found and write_if_missing is False.
        If ``write_if_missing`` is True a primitive cell is written to the database
        if it is missing.

        Returns the ID (an integer) of the row which corresponds to the primitive cell.
        """
        with self.connect() as con:
            try:
                # Check if the primitive has already been written to the database
                uid = _get_prim_cell_id_from_connection(self.prim_cell, con)
            except PrimitiveCellNotFound:
                # Primitive wasn't found
                if write_if_missing:
                    # Write it to the database.
                    uid = con.write(self.prim_cell, name="primitive_cell")
                else:
                    # We're not allowed to write to the database. Raise the error.
                    raise
        # Ensure connection is closed before returning, to ensure that the primitive
        # has been written.
        return uid

    @property
    def trans_matrix(self) -> TransMatrix:
        """Get the translation matrix, will be created upon request"""
        self.ensure_clusters_exist()
        return self._trans_matrix

    @property
    def cluster_list(self) -> ClusterList:
        """Get the cluster list, will be created upon request"""
        self.ensure_clusters_exist()
        return self._cluster_list

    def clear_cache(self) -> None:
        """Clear the cached objects, due to a change e.g. in the template atoms"""
        logger.debug("Clearing the cache")
        self._trans_matrix = None
        self._cluster_list = None

    def create_cluster_list_and_trans_matrix(self):
        """Prepares the internal cache objects by calculating cluster related properties"""
        logger.debug("Creating translation matrix and cluster list")
        self._cluster_list, self._trans_matrix = self.cluster_mng.build_all(
            self.atoms,
            self.max_cluster_dia,
            self.index_by_sublattice,
        )

    def view_clusters(self) -> None:
        """Display all clusters along with their names."""
        # pylint: disable=import-outside-toplevel
        from ase.gui.gui import GUI
        from ase.gui.images import Images

        figures = self.get_all_figures_as_atoms()
        images = Images()
        images.initialize(figures)
        gui = GUI(images, expr="")
        gui.show_name = True
        gui.run()

    def get_all_figures_as_atoms(self) -> List[Atoms]:
        """Get the list of all possible figures, in their
        ASE Atoms representation."""
        self.ensure_clusters_exist()
        return self.cluster_mng.get_figures()

    def requires_build(self) -> bool:
        """Check if the cluster list and trans matrix exist.
        Returns True the cluster list and trans matrix needs to be built.
        """
        return self._cluster_list is None or self._trans_matrix is None

    def ensure_clusters_exist(self) -> None:
        """Ensure the cluster list and trans matrix has been populated.
        They are not calculated upon creaton of the settings instance,
        for performance reasons. They will be constructed if required.
        Nothing is done if the cache exists."""
        if self.requires_build():
            logger.debug("Triggered construction of clusters")
            self.create_cluster_list_and_trans_matrix()
            # It should not be possible for them to be None after this call.
            assert self._cluster_list is not None
            assert self._trans_matrix is not None

    def get_all_templates(self):
        """
        Return a list with all template atoms
        """
        return self.template_atoms.get_all_templates()

    def view_templates(self):
        """
        Display all templates in the ASE GUi
        """
        # pylint: disable=import-outside-toplevel
        from ase.visualize import view

        view(self.get_all_templates())

    def _check_first_elements(self):
        basis_elements = self.basis_elements
        num_basis = self.num_basis
        # This condition can be relaxed in the future
        first_elements = []
        for elements in basis_elements:
            first_elements.append(elements[0])
        if len(set(first_elements)) != num_basis:
            raise ValueError("First element of different basis should not be the same.")

    def todict(self) -> Dict:
        """Return a dictionary representation of the settings class.

        Example:

            >>> from clease.settings import CEBulk, Concentration
            >>> conc = Concentration([['Au', 'Cu']])
            >>> settings = CEBulk(conc, crystalstructure='fcc', a=4.1)
            >>> dct = settings.todict()  # Get the dictionary representation
        """
        vars_to_save = self.ARG_KEYS + self.KWARG_KEYS + self.OTHER_KEYS
        dct = {"clease_version": str(__version__)}
        for key in vars_to_save:
            val = getattr(self, key)
            dct[key] = val
        return dct

    @classmethod
    def from_dict(cls, dct: Dict[str, Any]) -> ClusterExpansionSettings:
        """Load a new ClusterExpansionSettings class from a dictionary representation.

        Example:

            >>> from clease.settings import CEBulk, Concentration, ClusterExpansionSettings
            >>> conc = Concentration([['Au', 'Cu']])
            >>> settings = CEBulk(conc, crystalstructure='fcc', a=4.1)
            >>> dct = settings.todict()  # Get the dictionary representation
            >>> # Remove the existing settings, perhaps due to being in a new environment
            >>> del settings
            >>> # Load in the settins from the dictionary representation
            >>> settings = ClusterExpansionSettings.from_dict(dct)
        """
        dct = deepcopy(dct)

        # pop out the version. We can use this for compatibility checks if needed.
        dct.pop("clease_version", None)

        # Get the args and kwargs we expect in our function signature
        args = [dct.pop(key) for key in cls.ARG_KEYS]

        # Should we allow for missing kwargs keys?
        kwargs = {key: dct.pop(key) for key in cls.KWARG_KEYS}

        if "max_cluster_size" in dct:
            # For compatibility, since it's no longer in KWARG_KEYS
            # Avoid crashing.
            dct.pop("max_cluster_size")

        settings = cls(*args, **kwargs)

        for key in cls.OTHER_KEYS:
            # Populate "other" keys, which aren't *args or **kwargs
            value = dct.pop(key)
            setattr(settings, key, value)

        if dct:
            # We have some unexpected left-overs
            logger.warning("Unused items from dictionary: %s", dct)
        return settings

    def clusters_table(self) -> str:
        """String with information about the clusters"""
        # We need the clusters to be constructed first.
        self.ensure_clusters_exist()
        mult_dict = self.multiplicity_factor

        columns = [
            "Index",
            "Cluster Name",
            "Size",
            "Group",
            "Radius",
            "Figures",
            "Multiplicity",
        ]

        fmt = "| {:<5} | {:<12} | {:<4} | {:<5} | {:<6} | {:<7} | {:<12} |"
        header = fmt.format(*columns)
        rule = "-" * len(header)  # Horizontal line of -----
        lines = [rule, header, rule]

        for ii, cluster in enumerate(self.cluster_list.clusters):
            name = cluster.name
            size = f"{cluster.size:d}".center(4, " ")
            mult = f"{int(mult_dict[name]):d}".center(12, " ")
            radius = f"{cluster.diameter / 2:2.4f}"
            n_figures = f"{len(cluster.figures)}".center(7)
            group = f"{cluster.group:d}".center(5)
            index_s = f"{ii:d}".center(5)

            s = fmt.format(index_s, name, size, group, radius, n_figures, mult)
            lines.append(s)

        # Bottom horizontal line
        lines.append(rule)

        return "\n".join(lines)

    def get_cluster_corresponding_to_cf_name(self, cf_name: str) -> Cluster:
        """Find the Cluster object which corresponds to a CF name.
        The cluster will not be specialized to the decoration number if such exists
        in the cf name.

        Example:

            >>> from clease.settings import CEBulk, Concentration
            >>> conc = Concentration([['Au', 'Cu']])
            >>> settings = CEBulk(conc, crystalstructure='fcc', a=4.1)
            >>> cluster = settings.get_cluster_corresponding_to_cf_name("c1_0")
            >>> cluster.size
            1

        """
        size = get_size_from_cf_name(cf_name)
        for cluster in self.cluster_list.clusters:
            # This CF name may contain the decoration number, but the cluster name doesn't.
            # Decoration number is last, so test everything before that is equal.
            if cluster.size == size and cf_name.startswith(cluster.name):
                return cluster
        raise RuntimeError(f"Didn't find cluster corresponding to name: {cf_name}")


def _get_concentration(concentration: Union[Concentration, dict]) -> Concentration:
    """Helper function to format the concentration"""
    if isinstance(concentration, Concentration):
        conc = concentration
    elif isinstance(concentration, dict):
        conc = Concentration.from_dict(concentration)
    else:
        raise TypeError("concentration has to be either dict or instance of Concentration")
    return conc


def _get_prim_cell_id_from_connection(prim_cell: Atoms, connection: Database) -> int:
    """Retrieve the primitive cell ID from a database connection"""
    shape = prim_cell.cell.cellpar()
    for row in connection.select(name="primitive_cell"):
        loaded_shape = row.toatoms().cell.cellpar()
        if np.allclose(shape, loaded_shape):
            return row.id

    raise PrimitiveCellNotFound("The primitive cell was not found in the database.")



"""Module for generating new structures for training."""
import os
from copy import deepcopy
from functools import reduce
from typing import List, Dict, Optional, Union, Any, Tuple
import logging
from itertools import product

import numpy as np
from numpy.random import shuffle

from ase import Atoms
from ase.io import read
from ase.io.trajectory import TrajectoryReader
from ase.utils.structure_comparator import SymmetryEquivalenceCheck

from clease import db_util
from clease.settings import ClusterExpansionSettings
from clease.corr_func import CorrFunction

from clease.montecarlo import TooFewElementsError
from clease.tools import wrap_and_sort_by_position, nested_list2str
from clease.tools import count_atoms
from .structure_generator import ProbeStructure, GSStructure, MetropolisTrajectory

try:
    from math import gcd
except ImportError:
    from fractions import gcd

logger = logging.getLogger(__name__)

max_attempt = 10
max_fail = 10

__all__ = ("NewStructures", "MaxAttemptReachedError")


class NotValidTemplateException(Exception):
    """The template did not yield a valid atoms object"""


class MaxAttemptReachedError(Exception):
    """Raised when number of try reaches 10."""


class NewStructures:
    """
    Generate new structure in ASE Atoms object format.

    :param settings: Cluster expansion settings.

    :param generation_number: Generation number to be assigned to the newly
        generated structure

    :param struct_per_gen: Number of structures to generate per generation

    :param check_db: Should a new structure which is being inserted into the database be checked
        against pre-existing structures? Should only be disabled if you know what you are doing.
        Default is ``True``.
    """

    def __init__(
        self,
        settings: ClusterExpansionSettings,
        generation_number: int = None,
        struct_per_gen: int = 5,
        check_db: bool = True,
    ) -> None:
        self.check_db = check_db
        self.settings = settings
        self.corrfunc = CorrFunction(self.settings)
        self.struct_per_gen = struct_per_gen

        if generation_number is None:
            self.gen = self._determine_gen_number()
        else:
            self.gen = generation_number

    def connect(self, **kwargs):
        """Short-cut to access the settings connection."""
        return self.settings.connect(**kwargs)

    def num_in_gen(self) -> int:
        with self.connect() as con:
            cur = con.connection.cursor()
            cur.execute(
                "SELECT id FROM number_key_values WHERE key=? AND value=?",
                ("gen", self.gen),
            )
            num_in_gen = len(cur.fetchall())
        return num_in_gen

    def num_to_gen(self) -> int:
        return max(self.struct_per_gen - self.num_in_gen(), 0)

    def generate_probe_structure(
        self,
        atoms: Optional[Atoms] = None,
        init_temp: Optional[float] = None,
        final_temp: Optional[float] = None,
        num_temp: int = 5,
        num_steps_per_temp: int = 1000,
        approx_mean_var: bool = True,
        num_samples_var: int = 10000,
    ) -> None:
        """
        Generate a probe structure according to PRB 80, 165122 (2009).

        :param atoms: ASE Atoms object with the desired cell size and shape of
            the new structure.
        :param init_temp: initial temperature (does not represent *physical*
            temperature)
        :param final_temp: final temperature (does not represent *physical*
            temperature)
        :param num_temp: number of temperatures to be used in simulated
            annealing
        :param num_steps_per_temp: number of steps in simulated annealing
        :param approx_mean_var: whether or not to use a spherical and
            isotropical distribution approximation scheme for determining the
            mean variance.

            -`True`: Assume a spherical and isotropical distribution of
                structures in the configurational space. Corresponds to eq.4
                in PRB 80, 165122 (2009)
            -`False`: Use sigma and mu of eq.3 in PRB 80, 165122 (2009)
                to characterize the distribution of structures in population.
                Requires pre-sampling of random structures before generating
                probe structures. sigma and mu are generated and stored in
                'probe_structure-sigma_mu.npz' file.
        :param num_samples_var: number of samples to be used in determining
            signam and mu. Only used when `approx_mean_var` is `True`.

        Note: init_temp and final_temp are automatically generated if either
              one of the two is not specified.
        """

        if not approx_mean_var:
            # check to see if there are files containing mu and sigma values
            if not os.path.isfile("probe_structure-sigma_mu.npz"):
                self._generate_sigma_mu(num_samples_var)

        logger.info(
            ("Generate %s probe structures (generation: %s, struct_per_gen=%s, %s present)"),
            self.num_to_gen(),
            self.gen,
            self.struct_per_gen,
            self.num_in_gen(),
        )

        current_count = 0
        num_attempt = 0
        num_to_generate = self.num_to_gen()

        while current_count < num_to_generate:
            self.settings.set_active_template(atoms=atoms)
            # Break out of the loop if reached struct_per_gen
            num_struct = self.num_in_gen()
            if num_struct >= self.struct_per_gen:
                break

            struct = self._get_struct_at_conc(conc_type="random")

            logger.info("Generating structure %s out of %s.", current_count + 1, num_to_generate)
            ps = ProbeStructure(
                self.settings,
                struct,
                init_temp,
                final_temp,
                num_temp,
                num_steps_per_temp,
                approx_mean_var,
            )
            probe_struct, cf = ps.generate()

            # Remove energy from result dictionary
            probe_struct.calc.results.pop("energy")
            formula_unit = self._get_formula_unit(probe_struct)
            if self._exists_in_db(probe_struct, formula_unit):
                msg = "generated structure is already in DB.\n"
                msg += "generating again... "
                msg += f"{num_attempt + 1} out of {max_attempt} attempts."
                logger.info(msg)
                num_attempt += 1
                if num_attempt >= max_attempt:
                    msg = "Could not generate probe structure in "
                    msg += f"{max_attempt} attempts."
                    logger.error(msg)
                    raise MaxAttemptReachedError(msg)
            else:
                num_attempt = 0

            logger.info("Probe structure generated.")
            self.insert_structure(init_struct=probe_struct, cf=cf)
            current_count += 1

    @property
    def corr_func_table_name(self) -> str:
        return f"{self.settings.basis_func_type.name}_cf"

    def generate_gs_structure_multiple_templates(
        self,
        eci: Dict[str, float],
        num_templates: int = 20,
        num_prim_cells: int = 2,
        init_temp: float = 2000.0,
        final_temp: float = 1.0,
        num_temp: int = 10,
        num_steps_per_temp: int = 1000,
    ) -> None:
        """
        Generate ground-state structures using multiple templates
        (rather than using fixed cell size and shape). Structures are generated
        until the number of structures with the current `generation_number` in
        database reaches `struct_per_gen`.

        :param num_templates: Number of templates to search in. Simmulated
            annealing is done in each cell and the one with the lowest energy
            is taken as the ground state.
        :param num_prim_cells: Number of primitive cells to use when
            constructing templates. The volume of all the templates used will
            be num_prim_cells*vol_primitive, where vol_primitive is the volume
            of the primitive cell.

        See docstring of `generate_gs_structure` for the rest of the arguments.
        """

        for i in range(self.num_to_gen()):
            logger.info("Generating ground-state structures: %s of %s", i, self.struct_per_gen)

            self._generate_one_gs_structure_multiple_templates(
                eci=eci,
                num_templates=num_templates,
                num_prim_cells=num_prim_cells,
                init_temp=init_temp,
                final_temp=final_temp,
                num_temp=num_temp,
                num_steps_per_temp=num_steps_per_temp,
            )

    def _generate_one_gs_structure_multiple_templates(
        self,
        eci: Dict[str, float],
        num_templates: int = 20,
        num_prim_cells: int = 2,
        init_temp: float = 2000.0,
        final_temp: float = 1.0,
        num_temp: int = 10,
        num_steps_per_temp=1000,
    ) -> None:
        """
        Generate one ground-state structures using multiple templates
        (rather than using fixed cell size and shape).

        :param num_templates: Number of templates to search in. Simmulated
            annealing is done in each cell and the one with the lowest energy
            is taken as the ground state.
        :param num_prim_cells: Number of primitive cells to use when
            constructing templates. The volume of all the templates used will
            be num_prim_cells*vol_primitive, where vol_primitive is the volume
            of the primitive cell.

        See docstring of `generate_gs_structure` for the rest of the arguments.
        """
        templates = self.settings.template_atoms.get_fixed_volume_templates(
            num_templates=num_templates, num_prim_cells=num_prim_cells
        )

        if len(templates) == 0:
            msg = "Could not find any templates with matching the constraints"
            raise RuntimeError(msg)

        self.settings.set_active_template(atoms=templates[0])

        nib = [len(x) for x in self.settings.index_by_basis]
        x = self.settings.concentration.get_random_concentration(nib=nib)
        num_insert = self.settings.concentration.conc_in_int(nib, x)

        energies = []
        gs_structs = []
        cf_dicts = []
        for i, atoms in enumerate(templates):
            logger.info("Searching for GS in template %d of %d", i, len(templates))
            self.settings.set_active_template(atoms=atoms)

            struct = self._random_struct_at_conc(num_insert)
            es = GSStructure(
                self.settings,
                struct,
                init_temp,
                final_temp,
                num_temp,
                num_steps_per_temp,
                eci,
            )

            gs_struct, cf_struct = es.generate()
            gs_structs.append(gs_struct)
            cf_dicts.append(cf_struct)
            energies.append(gs_struct.get_potential_energy())

        # Find the position of the minimum energy structure
        min_energy_indx = np.argmin(energies)
        gs = gs_structs[min_energy_indx]
        cf = cf_dicts[min_energy_indx]

        self.settings.set_active_template(atoms=gs)
        self.insert_structure(init_struct=gs, cf=cf)

    def generate_gs_structure(
        self,
        atoms: Union[Atoms, List[Atoms]],
        eci: Dict[str, float],
        init_temp: float = 2000.0,
        final_temp: float = 1.0,
        num_temp: int = 10,
        num_steps_per_temp: int = 1000,
        random_composition: bool = False,
    ) -> None:
        """
        Generate ground-state structures based on cell sizes and shapes of the
        passed ASE Atoms.

        :param atoms: Atoms object with the desired size and composition of
            the new structure. A list of Atoms with different size and/or
            compositions can be passed. Compositions of the supplied Atoms
            object(s) are ignored when random_composition=True.
        :param eci: cluster names and their ECI values
        :param init_temp: Initial temperature (does not represent *physical*
            temperature)
        :param final_temp: Final temperature (does not represent *physical*
            temperature)
        :param num_temp: Number of temperatures to use in simulated annealing
        :param num_steps_per_temp: Number of steps in simulated annealing
        :param random_composition: Whether or not to fix the composition of the
            generated structure.

            1. *False* and atoms = Atoms object: One ground-state structure with
                matching size and composition of the supplied Atoms object is
                generated
            2. *False* and atoms = list: The same number of ground-state
                structures that matches the length of the list is generated

                * Note 1: num_struct_per_gen is ignored and all of the generated
                    structures have the same generation number
                * Note 2: each GS structure will have matching size and
                    composition of the suplied Atoms objects
            3. *True* and atoms = Atoms object: GS structure(s) with a
                matching size of the Atoms object is generated at a random
                composition (within the composition range specified in
                Concentration class)

                * Note 1: This will generate GS structures until the number of
                    structures with the current generation number equals
                    num_struct_per_gen
                * Note 2: A check is performed to ensure that none of the newly
                    generated GS structures have the same composition
            4. *True* and atoms = list: The same number of GS structures that
                matches the length of the list is generated

                * Note 1: num_struct_per_gen is ignored and all of the generated
                        structures have the same generation number
                * Note 2: each GS structure will have matching sizes of the
                        supplied Atoms objects but with a random composition
                * Note 3: No check is performed to ensure that all new GS
                        structures have unique composition
        """
        structs = self._set_initial_structures(atoms, random_composition)
        current_count = 0
        num_attempt = 0
        num_to_generate = min([self.num_to_gen(), len(structs)])
        while current_count < num_to_generate:
            struct = structs[current_count].copy()
            self.settings.set_active_template(atoms=struct)
            logger.info("Generating structure %d out of %d.", current_count + 1, num_to_generate)
            es = GSStructure(
                self.settings,
                struct,
                init_temp,
                final_temp,
                num_temp,
                num_steps_per_temp,
                eci,
            )
            gs_struct, cf = es.generate()
            formula_unit = self._get_formula_unit(gs_struct)

            if self._exists_in_db(gs_struct, formula_unit):
                msg = "generated structure is already in DB.\n"
                msg += "generating again... "
                msg += f"{num_attempt + 1} out of {max_attempt} attempts"
                logger.info(msg)
                num_attempt += 1
                if num_attempt >= max_attempt:
                    msg = f"Could not generate ground-state structure in {max_attempt} attempts."
                    logger.error(msg)
                    raise MaxAttemptReachedError(msg)
                continue

            num_attempt = 0

            min_energy = gs_struct.get_potential_energy()
            logger.info("Structure with E = %.3f generated.", min_energy)
            self.insert_structure(init_struct=gs_struct, cf=cf)
            current_count += 1

    def generate_random_structures(self, atoms: Optional[Atoms] = None) -> None:
        """
        Generate random structures until the number of structures with
        `generation_number` equals `struct_per_gen`.

        :param atoms: If Atoms object is passed, the passed object will be
            used as a template for all the random structures being generated.
            If None, a random template will be chosen.
            (different for each structure)
        """
        logger.info(
            ("Generating %d random structures (generation: %s, struct_per_gen=%d, %d present)."),
            self.num_to_gen(),
            self.gen,
            self.struct_per_gen,
            self.num_in_gen(),
        )

        fail_counter = 0
        i = 0

        num_structs = self.num_to_gen()
        while i < num_structs and fail_counter < max_fail:
            if self.generate_one_random_structure(atoms=atoms):
                i += 1
                fail_counter = 0
                logger.debug("Generated %d random structures", i)
            else:
                fail_counter += 1
                logger.debug("Failed generating structure. Fail count: %d", fail_counter)

        if fail_counter >= max_fail:
            msg = (
                "Could not find a structure that does not exist in DB after "
                f"{int(max_attempt * max_fail)} attempts."
            )
            logger.error(msg)
            raise MaxAttemptReachedError(msg)
        logger.info("Succesfully generated %d random structures.", self.num_in_gen())

    def generate_one_random_structure(self, atoms: Optional[Atoms] = None) -> bool:
        """
        Generate and insert a random structure to database if a unique
        structure is found.

        Returns ``True`` if unique structure is found and inserted in DB,
        ``False`` otherwise.

        :param atoms: If Atoms object is passed, the passed object will be
            used as a template for all the random structures being generated.
            If None, a random template will be chosen.
            (different for each structure)
        """
        num_attempts = 0
        unique_structure_found = False

        while not unique_structure_found and num_attempts < max_attempt:
            self.settings.set_active_template(atoms=atoms)
            new_atoms = self._get_struct_at_conc(conc_type="random")
            fu = self._get_formula_unit(new_atoms)
            if not self._exists_in_db(new_atoms, fu):
                unique_structure_found = True
            else:
                num_attempts += 1

        if not unique_structure_found:
            logger.warning(
                (
                    "Could not find a structure that does not already exist "
                    "in the DB within %d attempts."
                ),
                max_attempt,
            )
            return False

        logger.debug("Found unique structure after %d attempts.", num_attempts)
        self.insert_structure(init_struct=new_atoms)
        return True

    # pylint: disable=too-many-branches
    def _set_initial_structures(
        self, atoms: Union[Atoms, List[Atoms]], random_composition: bool = False
    ) -> List[Atoms]:
        structs = []
        if isinstance(atoms, Atoms):
            struct = wrap_and_sort_by_position(atoms)
            if random_composition is False:
                num_to_gen = 1
                logger.debug("Generate 1 ground-state structure.")
                structs.append(struct)
            else:
                msg = f"Generate {self.num_to_gen()} ground-state structures"
                msg += f"(generation: {self.gen}, "
                msg += f"struct_per_gen={self.struct_per_gen}, "
                msg += f"{self.num_in_gen()} present)"
                logger.info(msg)
                self.settings.set_active_template(atoms=struct)
                num_to_gen = self.num_to_gen()
                concs = []
                # Get unique concentrations
                num_attempt = 0
                nib = [len(x) for x in self.settings.index_by_basis]
                while len(concs) < num_to_gen:
                    x = self.settings.concentration.get_random_concentration(nib=nib)
                    if True in [np.allclose(x, i) for i in concs]:
                        num_attempt += 1
                    else:
                        concs.append(x)
                        num_attempt = 0

                    if num_attempt > 100:
                        msg = f"Could not find {self.num_to_gen()} unique "
                        msg += "compositions using the provided Atoms object."
                        raise RuntimeError(msg)
                num_atoms_in_basis = [len(indices) for indices in self.settings.index_by_basis]
                for x in concs:
                    num_insert = self.settings.concentration.conc_in_int(num_atoms_in_basis, x)
                    structs.append(self._random_struct_at_conc(num_insert))

        else:
            # This is a list of atoms objects
            logger.info("Generating %d ground-state structures.", len(atoms))
            if random_composition is False:
                for struct in atoms:
                    structs.append(wrap_and_sort_by_position(struct))
            else:
                concs = []
                nib = [len(x) for x in self.settings.index_by_basis]
                for struct in atoms:
                    self.settings.set_active_template(atoms=struct)
                    x = self.settings.concentration.get_random_concentration(nib=nib)
                    num_atoms_in_basis = [len(indices) for indices in self.settings.index_by_basis]
                    num_insert = self.settings.concentration.conc_in_int(num_atoms_in_basis, x)
                    structs.append(self._random_struct_at_conc(num_insert))
        return structs

    def generate_initial_pool(self, atoms: Optional[Atoms] = None) -> None:
        """
        Generate initial pool of structures.

        Initial pool of structures are generated, in sequence, using

        1. generate_conc_extrema(): structures at concentration where the
           number of consituting elements is at its max/min.
        2. generate_random_structures(): random structures are random
           concentration.

        Structures are genereated until the number of structures reaches
        `struct_per_gen`.

        :param atoms: If Atoms object is passed, the size and shape of its
                      cell will be used for all the random structures.
                      If None, a randome size and shape will be chosen for
                      each structure.
        """

        self.generate_conc_extrema()
        self.generate_random_structures(atoms=atoms)

    def generate_conc_extrema(self) -> None:
        """Generate initial pool of structures with max/min concentration."""
        logger.info(
            (
                "Generating one structure per concentration where the number "
                "of an element is at max/min"
            )
        )
        indx_in_each_basis = []
        start = 0
        for basis in self.settings.concentration.basis_elements:
            indx_in_each_basis.append(list(range(start, start + len(basis))))
            start += len(basis)

        for indx in product(*indx_in_each_basis):
            # We only iterate 1 template per size, as any other template with the same size
            # would only be able to accomodate the same concentrations,
            # so we need to increase the size anyway in case of an error.
            for template in self.settings.template_atoms.iterate_all_templates(max_per_size=1):
                self.settings.set_active_template(template)
                try:
                    atoms = self._get_struct_at_conc(conc_type="max", index=indx)
                except NotValidTemplateException:
                    continue
                atoms = wrap_and_sort_by_position(atoms)
                self.insert_structure(init_struct=atoms)
                break
            else:
                raise RuntimeError(f"Did not find a valid template for index {indx}")

    def generate_metropolis_trajectory(
        self, atoms: Optional[Atoms] = None, random_comp: bool = True
    ) -> None:
        """
        Generate a set of structures consists of single atom swaps

        :param atoms: ASE Atoms object that will be used as a template for the
            trajectory
        :param random_comp: If 'True' the passed atoms object will be
            initialised with a random composition. Otherwise, the trajectory
            will start from the passed Atoms object.
        """
        if atoms is None:
            atoms = self.settings.atoms.copy()

        if random_comp:
            self.settings.set_active_template(atoms=atoms)
            atoms = self._get_struct_at_conc(conc_type="random")

        num = self.num_to_gen()
        try:
            generator = MetropolisTrajectory(self.settings, atoms, num)
            all_atoms = generator.generate()
            for a in all_atoms:
                self.insert_structure(init_struct=a)

        except TooFewElementsError as exc:
            if random_comp:
                self.generate_metropolis_trajectory(atoms=atoms, random_comp=True)
            else:
                raise exc

    def _get_struct_at_conc(self, conc_type: str = "random", index: int = 0) -> Atoms:
        """Generate a structure at a concentration specified.

        :param conc_type: One of 'min', 'max' and 'random'
        :param index: Index of the flattened basis_element array to specify
            which element to be maximized/minimized
        """
        conc = self.settings.concentration
        if conc_type == "min":
            x = conc.get_conc_min_component(index)
        elif conc_type == "max":
            x = conc.get_conc_max_component(index)
        else:
            nib = [len(x) for x in self.settings.index_by_basis]
            x = conc.get_random_concentration(nib=nib)

        num_atoms_in_basis = [len(indices) for indices in self.settings.index_by_basis]
        num_to_insert = conc.conc_in_int(num_atoms_in_basis, x)
        atoms = self._random_struct_at_conc(num_to_insert)

        if conc_type in ["min", "max"]:
            # Check how close we got, and see if we got to an acceptable range
            new_conc = conc.get_concentration_vector(self.settings.index_by_basis, atoms)
            # check if we're close enough
            # TODO: Allow a tolerance here
            if not np.allclose(new_conc, x):
                raise NotValidTemplateException(
                    ("Did not find an atoms with a satisfactory concentration.")
                )

        return atoms

    def insert_structures(
        self, traj_init: str, traj_final: Optional[str] = None, cb=lambda num, tot: None
    ) -> None:
        """
        Insert a sequence of initial and final structures from their
        trajectory files.

        :param traj_init: Name of a trajectory file with initial structures
        :param traj_final: Name of a trajectory file with the final structures
        :param cb: Callback function that is called every time a structure is
            inserted (or rejected because it exists before). The signature of
            the function is cb(num, tot) where num is the number of inserted
            structure and tot is the total number of structures that should
            be inserted
        """
        traj_in = TrajectoryReader(traj_init)

        if traj_final is None:
            for i, init in enumerate(traj_in):
                self.insert_structure(init_struct=init)
                cb(i + 1, len(traj_in))
            return

        traj_final = TrajectoryReader(traj_final)

        if len(traj_in) != len(traj_final):
            raise ValueError("Different number of structures in initial trajectory file and final.")

        num_ins = 0
        for init, final in zip(traj_in, traj_final):
            # Check that composition (except vacancies matches)
            count_init = count_atoms(init)
            count_final = count_atoms(final)
            for key, value in count_final.items():
                if key not in count_init:
                    raise ValueError("Final and initial structure contains different elements")

                if count_init[key] != value:
                    raise ValueError(
                        "Final and initial structure has different number of each species"
                    )

            self.insert_structure(init_struct=init, final_struct=final)
            num_ins += 1
            cb(num_ins, len(traj_in))

    def insert_structure(
        self,
        init_struct: Union[Atoms, str],
        final_struct: Optional[Union[Atoms, str]] = None,
        name: Optional[str] = None,
        cf: Optional[Dict[str, float]] = None,
        meta: Optional[Dict[str, Any]] = None,
        warn_on_skip: bool = True,
    ) -> Optional[Union[int, Tuple[int, int]]]:
        """Insert a structure to the database.

        Returns the ID of the initial structure which was inserted into the database.
        If a row for the final structure is also inserted, a tuple of (initial_id, final_id)
        is returned. If no structure was inserted, ``Ǹone`` is returned, instead.

        :param init_struct: Unrelaxed initial structure. If a string is passed,
            it should be the file name with .xyz, .cif or .traj extension.
        :param final_struct: (Optional) final structure that contains energy.
            It can be either ASE Atoms object or file name readable by ASE.
        :param name: (Optional) name of the DB entry if a custom name is to be
            used. If `None`, default naming convention will be used.
        :param cf: (Optional) full correlation function of the initial
            structure (correlation functions with zero values should also be
            included). If cf is given, the preprocessing of the init_structure
            is bypassed and the given cf is inserted in DB.
        :param meta: (Optional) Extra information which will be added to the key-value
            pair entries in the database.
        :param warn_on_skip: (Bool, optional) Toggle emitting a warning if a structure
            was not inserted due to having a symmetrically equivalent structure
            in the database. Defaults to true.
        """
        if name is not None:
            with self.connect() as con:
                cur = con.connection.cursor()
                cur.execute(
                    "SELECT id FROM text_key_values WHERE key=? and value=?",
                    ("name", name),
                )
                num = len(cur.fetchall())
            if num > 0:
                raise ValueError(f"Name: {name} already exists in DB!")

        if cf is None:
            if isinstance(init_struct, Atoms):
                init_struct = wrap_and_sort_by_position(init_struct)
            else:
                init_struct = wrap_and_sort_by_position(read(init_struct))
            cf = self.corrfunc.get_cf(init_struct)

        self.settings.set_active_template(atoms=init_struct)

        formula_unit = self._get_formula_unit(init_struct)
        if self._exists_in_db(init_struct, formula_unit):
            if warn_on_skip:
                # Only emit a warning if requested.
                logger.warning(
                    "Supplied structure already exists in DB. The structure will not be inserted."
                )
            return None

        kvp = self._get_kvp(formula_unit)

        if name is not None:
            kvp["name"] = name
        kvp["converged"] = False
        kvp["started"] = False
        kvp["queued"] = False
        kvp["struct_type"] = "initial"
        if meta is not None:
            # Add optional meta-data. Ensure we do not have conflicts.
            for key, value in meta.items():
                if key in kvp:
                    raise KeyError(f"Key {key} is reserved.")
                kvp[key] = value
        tab_name = self.corr_func_table_name
        con = self.connect()
        uid_init: int = db_util.new_row_with_single_table(con, init_struct, tab_name, cf, **kvp)
        ret_val = uid_init

        if final_struct is not None:
            if not isinstance(final_struct, Atoms):
                final_struct = read(final_struct)
            kvp_final = {"struct_type": "final", "name": kvp["name"]}
            uid_final: int = con.write(final_struct, kvp_final)
            con.update(uid_init, converged=True, started="", queued="", final_struct_id=uid_final)
            ret_val = (uid_init, uid_final)

        return ret_val

    def _exists_in_db(self, atoms: Atoms, formula_unit: Optional[str] = None) -> bool:
        """
        Check to see if the passed atoms already exists in DB.

        To reduce the number of assessments for symmetry equivalence,
        check is only performed with the entries with the same concentration
        value.

        Return *True* if there is a symmetry-equivalent structure in DB,
        return *False* otherwise.

        :param atoms: Structure to be compared against the rest of structures
            in DB.
        :formula_unit: Reduced formula unit of the passed Atoms object
        """
        if not self.check_db:
            # No comparison was requested, so short-circuit the symmetry check
            return False
        cond = [("name", "!=", "template"), ("name", "!=", "primitive_cell")]
        if formula_unit is not None:
            cond.append(("formula_unit", "=", formula_unit))

        to_prim = True
        try:
            __import__("spglib")
        except ImportError:
            logger.warning("Setting 'to_primitive=False' because spglib is missing!")
            to_prim = False

        symmcheck = SymmetryEquivalenceCheck(
            angle_tol=1.0, ltol=0.05, stol=0.05, scale_volume=True, to_primitive=to_prim
        )

        with self.connect() as con:
            atoms_in_db = [row.toatoms() for row in con.select(cond)]

        return symmcheck.compare(atoms.copy(), atoms_in_db)

    def _get_kvp(self, formula_unit: str = None) -> Dict:
        """
        Create a dictionary of key-value pairs and return it.

        :param atoms: ASE Atoms object for which the key-value pair
            descriptions will be generated
        :param formula_unit: reduced formula unit of the passed Atoms object
        """
        if formula_unit is None:
            raise ValueError("Formula unit not specified!")
        kvp = {}
        kvp["gen"] = self.gen
        kvp["converged"] = False
        kvp["started"] = False
        kvp["queued"] = False

        suffixes = []
        logger.debug("Connecting to %s", self.settings.db_name)
        with self.connect() as con:
            cur = con.connection.cursor()
            logger.debug("Selecting from db: formula_unit=%s", formula_unit)
            cur.execute(
                "SELECT id FROM text_key_values WHERE key=? AND value=?",
                ("formula_unit", formula_unit),
            )
            ids = [i[0] for i in cur.fetchall()]
            for row_id in ids:
                logger.debug("Selecting from db: name=%s", row_id)
                cur.execute(
                    "SELECT value FROM text_key_values WHERE key=? AND id=?",
                    ("name", row_id),
                )
                name = cur.fetchone()[0]
                suffix = 0
                if "_" in name:
                    suffix = int(name.rpartition("_")[-1])
                suffixes.append(suffix)
        suffixes.sort()

        suffix = len(suffixes)
        for i, s in enumerate(suffixes):
            if i != s and i not in suffixes:
                suffix = i
                break

        suffix = min(suffix, len(suffixes))
        kvp["name"] = formula_unit + f"_{suffix}"
        kvp["formula_unit"] = formula_unit
        kvp["struct_type"] = "initial"
        size = self.settings.size
        if size is not None:
            # We do not store the size if it is None
            size = nested_list2str(self.settings.size)
            kvp["size"] = size
        return kvp

    def _get_formula_unit(self, atoms: Atoms) -> str:
        """Generates a reduced formula unit for the structure."""
        atom_count = []
        all_nums = []
        for group in self.settings.index_by_basis:
            new_count = {}
            for indx in group:
                symbol = atoms[indx].symbol
                if symbol not in new_count:
                    new_count[symbol] = 1
                else:
                    new_count[symbol] += 1
            atom_count.append(new_count)
            all_nums += [v for k, v in new_count.items()]
        # pylint: disable=unnecessary-lambda
        gcdp = reduce(lambda x, y: gcd(x, y), all_nums)
        fu = ""
        for i, count in enumerate(atom_count):
            keys = list(count.keys())
            keys.sort()
            for k in keys:
                fu += f"{k}{int(count[k] / gcdp)}"
            if i < len(atom_count) - 1:
                fu += "_"
        return fu

    def _random_struct_at_conc(self, num_atoms_to_insert: np.ndarray) -> Atoms:
        """Generate a random structure."""
        rnd_indices = []
        for indices in self.settings.index_by_basis:
            rnd_indices.append(deepcopy(indices))
            shuffle(rnd_indices[-1])

        # Insert the number of atoms
        basis_elem = self.settings.concentration.basis_elements
        assert len(rnd_indices) == len(basis_elem)
        atoms = self.settings.atoms.copy()
        current_conc = 0
        num_atoms_inserted = 0

        for basis, indices in enumerate(rnd_indices):
            current_indx = 0
            for symb in basis_elem[basis]:
                for _ in range(num_atoms_to_insert[current_conc]):
                    atoms[indices[current_indx]].symbol = symb
                    current_indx += 1
                    num_atoms_inserted += 1
                current_conc += 1
        assert num_atoms_inserted == len(atoms)
        return atoms

    def _determine_gen_number(self) -> int:
        """Determine generation number based on the values in DB."""
        with self.connect() as con:
            cur = con.connection.cursor()
            cur.execute("SELECT value FROM number_key_values WHERE key='gen'")
            gens = [int(i[0]) for i in cur.fetchall()]
            if len(gens) == 0:
                gen = 0
            else:
                gen = max(gens)
                cur.execute(
                    "SELECT id FROM number_key_values WHERE key=? AND value=?",
                    ("gen", gen),
                )
                num_in_gen = len(cur.fetchall())
                if num_in_gen >= self.struct_per_gen:
                    gen += 1
        return gen

    def _generate_sigma_mu(self, num_samples_var: int) -> None:
        """
        Generate sigma and mu of eq.3 in PRB 80, 165122 (2009) and save them
        in `probe_structure-sigma_mu.npz` file.

        :param num_samples_var: number of samples to be used in determining
                                signam and mu.
        """
        logger.info(
            (
                "===========================================================\n"
                "Determining sigma and mu value for assessing mean variance.\n"
                "May take a long time depending on the number of samples \n"
                "specified in the *num_samples_var* argument, which is %d.\n"
                "==========================================================="
            ),
            num_samples_var,
        )
        count = 0
        cfm = np.zeros((num_samples_var, len(self.settings.all_cf_names)), dtype=float)
        while count < num_samples_var:
            atoms = self._get_struct_at_conc(conc_type="random")
            cfm[count] = self.corrfunc.get_cf(atoms)
            count += 1
            logger.info("Sampling %d ouf of %d", count, num_samples_var)

        sigma = np.cov(cfm.T)
        mu = np.mean(cfm, axis=0)
        fname = "probe_structure-sigma_mu.npz"  # Should probably be a variable
        np.savez(fname, sigma=sigma, mu=mu)
        logger.debug("Saved sigma and mu in %s", fname)



from __future__ import annotations
from typing import Iterator, List
import ase
from clease.datastructures import MCStep
from clease.calculator import CleaseCacheCalculator
from .mc_observer import MCObserver

__all__ = ("MoveObserver",)


class MoveObserver(MCObserver):
    """Store each step from an MC run to reconstruct the individual atoms objects later.

    The interval must be set to 1 when attaching this observer,
    as otherwise steps may be lost and the reconstruction may end up being incorrect.

    Args:
        base_atoms (ase.Atoms): The base atoms object which is run in the MC.
        only_accept (bool, optional): Selects whether the only accepted moves or
            all the attempted moves are saved. If False, every move will be saved.
            Defaults to False.
    """

    def __init__(self, base_atoms: ase.Atoms, only_accept: bool = False):
        super().__init__()
        self.base_atoms = base_atoms.copy()
        self.only_accept = only_accept

        self.steps = []

    def observe_step(self, mc_step: MCStep) -> None:
        """Observe a single step"""
        if self.only_accept is True and mc_step.move_accepted is False:
            # Step was rejected, and we only want accepted steps
            return
        self.steps.append(mc_step)

    def reset(self) -> None:
        self.steps = []

    def reconstruct(self) -> List[ase.Atoms]:
        """Rebuild the atoms objects as defined by the observed changes."""
        return list(self.reconstruct_iter())

    def reconstruct_iter(self) -> Iterator[ase.Atoms]:
        """Iterator which builds the atoms objects 1-by-1."""
        atoms = self.base_atoms.copy()
        for step in self.steps:
            # Update the current atoms
            atoms = _apply_step(atoms, step)
            yield atoms

    def interval_ok(self, interval: int) -> bool:
        """Missing steps will result in incorrect reconstructions"""
        return interval == 1


def _apply_step(atoms: ase.Atoms, step: MCStep) -> ase.Atoms:
    """Helper function to apply a step to an atoms object.
    Always a returns a copy of the atoms object."""
    atoms = atoms.copy()
    # Apply changes to a copy of the atoms
    if step.move_accepted:
        for change in step.last_move:
            change.apply_change(atoms)
    # Attach a calculator with the energies
    calc = CleaseCacheCalculator(energy=step.energy)
    atoms.calc = calc
    return atoms



"""Module for tools pertaining to geometry of atoms and cells."""
import numpy as np
from ase import Atoms

__all__ = ("max_sphere_dia_in_cell", "supercell_which_contains_sphere", "cell_wall_distances")


def supercell_which_contains_sphere(atoms: Atoms, diameter: float) -> Atoms:
    """Find the smallest supercell of an atoms object which can contain a sphere,
    using only repetitions of (nx, ny, nz) (i.e. a diagonal P matrix).

    The number of repetitions is stored in the info dictionary of the supercell under the
    "repeats" keyword, i.e. sc.info["repeats"] is the number of times the supercell was repeated.

    Args:
        atoms (Atoms): The atoms object to be enlarged.
        diameter (float): The diameter of the sphere which should be contained within the supercell.

    Returns:
        Atoms: The supercell which can contain a sphere of the given diameter.
    """
    cell = atoms.get_cell()
    dists = cell_wall_distances(cell)
    # Find the number of repetitions required for each cell
    # wall to be larger than the diameter, i.e. allowing
    # a sphere with the provided diameter to exist within the cell.
    reps = np.ceil(diameter / dists).astype(int)
    assert reps.shape == (3,)
    sc: Atoms = atoms * reps
    assert cell_wall_distances(sc.get_cell()).min() >= diameter

    sc.info["repeats"] = reps

    return sc


def cell_wall_distances(cell: np.ndarray) -> np.ndarray:
    """Get the distances from each cell wall to the opposite cell wall of the cell.
    Returns the distances in the order of the distances between the (b, c), (a, c) and (a, b)
    planes, such that the shorest vector corresponds to being limited by the a, b or c vector,
    respectively.

    Args:
        cell (np.ndarray): A (3 x 3) matrix which defines the cell parameters.
            Raises a ValueError if the cell shape is wrong.

    Returns:
        np.ndarray: The distance to each of the three cell walls.
    """
    if not cell.shape == (3, 3):
        raise ValueError(f"Cell should be (3, 3), got {cell.shape}")
    # Midpoint of the parallelogram spanned by the cell vectors
    midpoint = np.sum(cell / 2, axis=0)

    # Get the shortest distance between the midpoint and planes defined by the
    # ab, ac and bc vectors. Since the cell is a parallelogram, the 3 remaining planes will
    # have identical distances to the midpoint.
    # Assumption: One of the cell corners is at (0, 0, 0),
    # so 3 of the 6 planes will share the point (0, 0, 0).
    # PQ is the vector from our midpoint to the reference point in the plane.
    PQ = midpoint - [0, 0, 0]
    a, b, c = cell

    def _iter_distances() -> float:
        # Iterate all distances from cell wall to the midpoint
        # The ordering here ensures that the distances corresponds to the length
        # a, b, c in a cubic cell (in that order)
        # Alternatively: The vector which limits the size of the sphere
        # is determined by the argmin of this array.
        for v1, v2 in [(b, c), (a, c), (a, b)]:
            # Find the normal vector to the plane spanned by v1 and v2
            # Note: The astype(float) is to ensure that a cell defined by
            # integers still works.
            n = np.cross(v1, v2).astype(float)
            n /= np.linalg.norm(n)
            # Project onto the normal vector to get the shortest possible
            # distance.
            # Multiply by 2 to get the diameter.
            yield np.linalg.norm(PQ.dot(n)) * 2

    return np.array(list(_iter_distances()), dtype=float)


def max_sphere_dia_in_cell(cell: np.ndarray) -> float:
    """Find the diameter of the largest possible sphere which can be placed
    inside a cell.

    For example, how large of a Death Star could be built inside a given atoms object?

    Args:
        cell (np.ndarray): A (3 x 3) matrix which defines the cell parameters.
            Raises a ValueError if the cell shape is wrong.

    Returns:
        float: The diameter of the largest sphere which can fit within the cell.
    """
    # The smallest possible distance is the radius of the largest sphere we can have.
    return cell_wall_distances(cell).min()



from typing import Union
from pathlib import Path
import numpy as np

__all__ = ("BufferedArray",)


class BufferedArray:
    """
    Implement a cyclic buffer. If fname, is given the buffer is flushed to
    file when full.

    :param size: Size of the buffer
    :param fname: Filename where the buffer will be written when full. If None,
        the buffer will not be flushed to file when full.
    """

    def __init__(self, size: int = 1000, fname: Union[str, Path] = None):
        self._buffer = np.zeros(size)
        self._next = 0
        self.fname = fname

    def flush(self):
        if self.fname is not None:
            with open(self.fname, "a") as f:
                np.savetxt(f, self._buffer)
        self._next = 0
        self._buffer[:] = 0

    def _flush_if_full(self) -> None:
        """
        Flush buffer to backup file when the buffer is full. If no file is
        given, this function simply resets the buffer to zeros.
        """
        if self._next < len(self._buffer):
            return
        self.flush()

    def push(self, v: float) -> None:
        """
        Insert a new value at the end of the buffer. The buffer is automatically
        flushed if full.

        :param v: New value to insert.
        """
        self._flush_if_full()
        self._buffer[self._next] = v
        self._next += 1

    def clear(self):
        """
        Clear array
        """
        self._buffer[:] = 0
        self._next = 0



import numpy as np
from ase.calculators.calculator import Calculator, PropertyNotImplementedError


class CleaseCacheCalculator(Calculator):
    """This is a simple clease calculator cache object,
    similar to the ASE singlepoint calculator, but less restrictive"""

    name = "clease_cache"

    def __init__(self, **results):
        super().__init__()
        self.results = {}

        self.results.update(**results)

    def __str__(self) -> str:
        tokens = []
        for key, val in sorted(self.results.items()):
            if np.isscalar(val):
                txt = f"{key}={val}"
            else:
                txt = f"{key}=..."
            tokens.append(txt)
        return "{}({})".format(self.__class__.__name__, ", ".join(tokens))

    def get_property(self, name, atoms=None, allow_calculation=True):
        if name not in self.results:
            raise PropertyNotImplementedError("The property '{name}' is not available.")

        result = self.results[name]
        if isinstance(result, np.ndarray):
            result = result.copy()
        return result

    def check_state(self, *args, **kwargs):
        """Will never fail the check_state check"""
        # pylint: disable=unused-argument
        return []



"""Expose the C++ Cluster object to Python"""
# distutils: language = c++


from cluster cimport Cluster

cdef class CppCluster:
    cdef Cluster *thisptr

    def __init__(self, cluster_obj):
        """Construct the C++ Cluster from a regular CLEASE
        python cluster object.

        Parameters:

        :cluster_obj: Instance of a clease.cluster.Cluster object
        """
        self.thisptr = new Cluster(cluster_obj)

    def __dealloc__(self):
        del self.thisptr

    def get_all_decoration_numbers(self, n_basis_funcs):
        return self.thisptr.get_all_decoration_numbers(n_basis_funcs)

    def get_size(self):
        return self.thisptr.get_size()



import click
from clease.settings import settings_from_json
from clease import tools
from . import main


@main.clease_cli.command()
@click.argument(
    "filename",
    type=click.Path(exists=True),
)
def reconfigure(filename):
    """
    Reconfigure a settings file saved as a JSON file.
    Will reconfigure the database that the settings points to.

    FILENAME is the name of the settings JSON file
    """
    settings = settings_from_json(filename)
    tools.reconfigure(settings, verbose=True)



from typing import Union, Sequence, Callable, List
from tkinter import TclError
import numpy as np
from ase.db import connect
from ase.gui.gui import GUI
from ase.gui.images import Images
import attr


@attr.define
class MPLEvent:
    """Simple conatiner for a matplotlib event"""

    event_name: str
    func: Callable[[], None]


@attr.define
class AnnotatedAx:
    """
    Container for collecting a matplotlib axis (ax) object along
    with annotations for each datapoint in the plot.

    ax: Axes instance
        Instance of the axes object containing the lines

    lines: Array of Line objects that annotations apply to

    annotations: Nested list with annotations
        Each list contains one annotation for each (x, y) pair on that
        line
    structure_names (Optional): Same shape as annotations, which contains the name of
        a structure for fetching data from the database.
    """

    ax = attr.field()
    lines = attr.field()
    annotations = attr.field()
    structure_names = attr.field(default=None)

    def __attrs_post_init__(self):
        for i, line in enumerate(self.lines):
            x, _ = line.get_data()
            if len(x) != len(self.annotations[i]):
                msg = f"Annotations for line {i} "
                msg += f"has length {len(self.annotations[i])} "
                msg += f"but there are {len(x)} data points!"
                raise ValueError(msg)


class InteractivePlot:
    """Interactive plot with annotations.

    Parameters:

    fig: Figure instance
        Instance of the figure object visualizing the data

    annotated_axes: Sequence of AnnotatedAx, which contains information
        on the annotations for lines in the ax object.
    """

    def __init__(self, fig, annotated_axes: Union[AnnotatedAx, Sequence[AnnotatedAx]]):
        self.fig = fig
        self.annotated_axes = annotated_axes

        self.all_annotations = []
        for annot_ax in self.annotated_axes:
            annot = annot_ax.ax.annotate(
                "",
                xy=(0, 0),
                xytext=(-20, 20),
                textcoords="offset points",
                bbox=dict(boxstyle="round", fc="w"),
                arrowprops=dict(arrowstyle="->"),
            )
            annot.set_visible(False)
            self.all_annotations.append(annot)
        assert len(self.all_annotations) == len(self.annotated_axes)
        # Initialize the active annotation to the first annotation object.
        # This reference may be moved to a different axis later.
        self.active_annot = self.all_annotations[0]
        self.active_line_index = 0

        self.connect_mpl()

    def get_mpl_events(self) -> List[MPLEvent]:
        event = MPLEvent("motion_notify_event", self.hover)
        return [event]

    def connect_mpl(self) -> None:
        """Connect all MPL events to the canvas.
        Ensure each event has a strong reference, so they are not garbage collected
        if this class falls out of scope."""
        for mpl_event in self.get_mpl_events():
            # Apply the event_wrapper to make a stronger reference. See note on event handling:
            # https://matplotlib.org/stable/users/explain/event_handling.html
            self.fig.canvas.mpl_connect(mpl_event.event_name, _event_wrapper(mpl_event.func))

    @property
    def annotated_axes(self) -> Sequence[AnnotatedAx]:
        return self._annotated_axes

    @annotated_axes.setter
    def annotated_axes(self, value) -> None:
        if isinstance(value, AnnotatedAx):
            # It's a single AnnotatedAx object, put it in a list
            # with itself, so the rest of the logic fits.
            self._annotated_axes = [value]
            return
        # Assume it's a list-like type of AnnotatedAx objects
        # Verify that each one is of the correct type.
        for annot_ax in value:
            if not isinstance(annot_ax, AnnotatedAx):
                raise TypeError(f"Each annotated axis object must be AnnotatedAx, got {annot_ax!r}")
        self._annotated_axes = value

    def _update_annotation(self, ind, event_index: int):
        """Update the annotation shown."""
        ax_obj = self.annotated_axes[event_index]
        ax = ax_obj.ax
        lines = ax_obj.lines
        line = lines[self.active_line_index]
        x, y = line.get_data()
        self.active_annot = self.all_annotations[event_index]
        xy = x[ind["ind"][0]], y[ind["ind"][0]]
        self.active_annot.xy = xy

        # Adjust the xytext coordinates away from the figure edges.
        # Calculate a vector towards the center of the plot
        xlim = ax.get_xlim()
        ylim = ax.get_ylim()

        center = np.array([sum(xlim), sum(ylim)]) / 2
        delta = center - xy

        # Set the values depending on the sign of the difference vector.
        # vector has constant length, of direction towards the center of the plot.
        # (i.e. away from the edges)
        # Values chosen by eye, but are scale-independent.
        sgn = np.sign(delta)
        xnew = 20 if sgn[0] >= 0 else -100
        ynew = 20 if sgn[1] >= 0 else -80

        self.active_annot.set_x(xnew)
        self.active_annot.set_y(ynew)

        anot = ax_obj.annotations[self.active_line_index]
        text = anot[ind["ind"][0]]
        self.active_annot.set_text(text)

    def hover(self, event):
        """React on a hover event."""
        vis = self.active_annot.get_visible()
        if self.event_in_ax(event):
            event_index = self.get_event_index(event)
            cont = False
            lines = self.annotated_axes[event_index].lines
            for i, line in enumerate(lines):
                cont, ind = line.contains(event)
                if cont:
                    self.active_line_index = i
                    break

            if cont:
                self._update_annotation(ind, event_index)
                self.active_annot.set_visible(True)

                self.fig.canvas.draw_idle()
            else:
                if vis:
                    self.active_annot.set_visible(False)
                    self.fig.canvas.draw_idle()

    def event_in_ax(self, event) -> bool:
        """Is an event within an active axes?"""
        return any(event.inaxes == annot_ax.ax for annot_ax in self.annotated_axes)

    def get_event_index(self, event) -> int:
        """Get the index of the axes object corresponding to the event."""
        for ii, annot_ax in enumerate(self.annotated_axes):
            if annot_ax.ax == event.inaxes:
                return ii
        raise RuntimeError(f"Didn't find index corresponding to event {event}.")

    def get_event_ax(self, event) -> AnnotatedAx:
        """Retrieve the AnnotatedAx object which corresponds to an event."""
        idx = self.get_event_index(event)
        return self.annotated_axes[idx]


class ShowStructureOnClick(InteractivePlot):
    def __init__(self, fig, axes: Union[AnnotatedAx, Sequence[AnnotatedAx]], db_name: str):
        self.db_name = db_name
        self.active_images = Images()
        # Uninitilized GUI. We don't create this until we click,
        # since we otherwise just open a new empty GUI window.
        self.gui = None
        super().__init__(fig, axes)

    def get_mpl_events(self) -> List[MPLEvent]:
        event = MPLEvent("button_press_event", self.on_click)
        events = super().get_mpl_events()
        events.append(event)
        return events

    def on_click(self, event) -> None:

        if not self.event_in_ax(event):
            return

        if event.button == 1:
            # Find the index of the point
            annot_ax = self.get_event_ax(event)

            for i, line in enumerate(annot_ax.lines):
                is_contained, ind = line.contains(event)
                if is_contained:
                    self.active_line_index = i
                    break
            else:
                # We clicked on something which wasn't a datapoint
                # because we never broke out of the loop.
                return

            # Prepare to grab the atoms object from the data base.
            db = connect(self.db_name)
            atoms = []
            self._close_gui()

            if annot_ax.structure_names is not None:
                name = annot_ax.structure_names[self.active_line_index][ind["ind"][0]]
            else:
                # Try falling back on the annotations as name (old)
                name = annot_ax.annotations[self.active_line_index][ind["ind"][0]]

            for row in db.select(name=name):
                atoms.append(row.toatoms())
            if not atoms:
                # Didn't find anything.
                return
            self.active_images.initialize(atoms)
            # Create a new GUI instance and run it
            self.gui = GUI(self.active_images)
            self.gui.run()

    def _close_gui(self) -> None:
        if self.gui is not None:
            # Try to close the existing GUI, if we already opened one
            try:
                self.gui.exit()
            except TclError:
                pass


def _event_wrapper(fnc):
    """Wrapper for an event function, to avoid a weak-referenced
    method is garbage collected. Otherwise, the weak reference may be lost,
    and the event is removed from the matplotlib callback."""

    def _wrapper(*args, **kwargs):
        fnc(*args, **kwargs)

    return _wrapper



"""Module for calculating correlation functions."""
from typing import Iterator, Tuple, Dict, Any
import logging
from ase.atoms import Atoms

from clease_cxx import PyCEUpdater
from .settings import ClusterExpansionSettings
from .tools import wrap_and_sort_by_position
from . import db_util

logger = logging.getLogger(__name__)
__all__ = ("CorrFunction", "ClusterNotTrackedError")

# Type alias for a Correlation function
CF_T = Dict[str, float]


class ClusterNotTrackedError(Exception):
    """A cluster is not being tracked"""


class CorrFunction:
    """Class for calculating the correlation functions.

    Parameters:

        settings (ClusterExpansionSettings): The settings object which defines the
            cluster expansion parameters.
    """

    def __init__(self, settings: ClusterExpansionSettings):
        self.settings = settings

    @property
    def settings(self) -> ClusterExpansionSettings:
        return self._settings

    @settings.setter
    def settings(self, value: Any) -> None:
        if not isinstance(value, ClusterExpansionSettings):
            raise TypeError(f"Setting must be a ClusterExpansionSettings object, got {value!r}")
        self._settings = value

    def connect(self, **kwargs):
        return self.settings.connect(**kwargs)

    def get_cf(self, atoms) -> CF_T:
        """
        Calculate correlation functions for all possible clusters and return
        them in a dictionary format.

        Parameters:

            atoms (Atoms): The atoms object
        """
        if not isinstance(atoms, Atoms):
            raise TypeError("atoms must be an Atoms object")
        cf_names = self.settings.all_cf_names
        return self.get_cf_by_names(atoms, cf_names)

    def get_cf_by_names(self, atoms, cf_names) -> CF_T:
        """
        Calculate correlation functions of the specified clusters and return
        them in a dictionary format.

        Parameters:

            atoms: Atoms object

            cf_names: list
                names of correlation functions that will be calculated for
                the structure provided in atoms
        """

        if isinstance(atoms, Atoms):
            self.set_template(atoms)
        else:
            raise TypeError("atoms must be Atoms object")

        self._confirm_cf_names_exists(cf_names)

        eci = {name: 1.0 for name in cf_names}
        cf = {name: 1.0 for name in cf_names}
        updater = PyCEUpdater(atoms, self.settings, cf, eci, self.settings.cluster_list)
        cf = updater.calculate_cf_from_scratch(atoms, cf_names)
        return cf

    def reconfigure_single_db_entry(self, row_id: int) -> None:
        """Reconfigure a single DB entry. Assumes this is the initial structure,
        and will not check that.

        Parameters:

            row_id: int
                The ID of the row to be reconfigured.
        """
        with self.connect() as db:
            atoms = wrap_and_sort_by_position(db.get(id=row_id).toatoms())
            cf = self.get_cf(atoms)
            db_util.update_table(db, row_id, self.cf_table_name, cf)

    @property
    def cf_table_name(self) -> str:
        """Name of the table which holds the correlation functions."""
        return f"{self.settings.basis_func_type.name}_cf"

    def clear_cf_table(self) -> None:
        """Delete the external table which holds the correlation functions."""
        with self.connect() as db:
            db.delete_external_table(self.cf_table_name)

    def reconfigure_db_entries(self, select_cond=None, verbose=False):
        """Reconfigure the correlation function values of the entries in DB.

        Parameters:

            select_cond: One of either:

                - None (default): select every item in DB with ``struct_type='initial'``
                - Select based on the condictions provided (``struct_type='initial'`` is
                  not automatically included)

            verbose (bool):
                print the progress of reconfiguration if set to *True*
        """
        select = format_selection(select_cond)
        db = self.connect()

        # get how many entries need to be reconfigured
        num_reconf = db.count(select)
        msg = f"{num_reconf} entries will be reconfigured"
        logger.info(msg)
        if verbose:
            print(msg)

        for row_id, count, total in self.iter_reconfigure_db_entries(select_cond=select_cond):
            msg = f"Updated {count} of {total} entries. Current ID: {row_id}"
            if verbose:
                print(msg, end="\r")
            logger.debug(msg)

        if verbose:
            print("\nreconfiguration completed")
        logger.info("Reconfiguration complete")

    def iter_reconfigure_db_entries(self, select_cond=None) -> Iterator[Tuple[int, int, int]]:
        """Iterator which reconfigures the correlation function values in the DB,
        which yields after each reconfiguration and reports on the progress.

        For more information, see :py:meth:`~reconfigure_db_entries`.

        Yields:
            Tuple[int, int, int]: (row_id, count, total) A tuple containing the ID
            of the row which was just reconfigured, current
            count which has been reconfigured, as well as the total number of
            reconfigurations which will be performed.
            The percentage-wise progress is thus (count / total) * 100.
        """
        # Setup, ensure CF's are cleared first
        self.clear_cf_table()
        db = self.connect()
        select = format_selection(select_cond)
        # Fetch the total number of reconfigures
        total = db.count(select)
        for count, row in enumerate(db.select(select), start=1):
            row_id = row.id
            self.reconfigure_single_db_entry(row_id)
            # Yield the current progress
            yield row_id, count, total

    def reconfigure_inconsistent_cf_table_entries(self):
        """Find and correct inconsistent correlation functions in table."""
        inconsistent_ids = self.check_consistency_of_cf_table_entries()

        if len(inconsistent_ids) == 0:
            return

        logger.info("Reconfiguring correlation functions")
        for count, bad_id in enumerate(inconsistent_ids):
            logger.debug(
                "Updating %s of %s entries (id %s)",
                count + 1,
                len(inconsistent_ids),
                bad_id,
            )
            self.reconfigure_db_entries(select_cond=[("id", "=", bad_id)], verbose=False)
        logger.info("Reconfiguration completed")

    def check_consistency_of_cf_table_entries(self):
        """Get IDs of the structures with inconsistent correlation functions.

        Note: consisent structures have the exactly the same list of cluster
        names as stored in settings.cf_names.
        """
        db = self.connect()
        tab_name = self.cf_table_name
        cf_names = sorted(self.settings.all_cf_names)
        inconsistent_ids = []
        for row in db.select("struct_type=initial"):
            tab_entries = row.get(tab_name, {})
            row_cnames = sorted(list(tab_entries.keys()))
            if row_cnames != cf_names:
                inconsistent_ids.append(row.id)

        if len(inconsistent_ids) > 0:
            logger.warning(
                "%d inconsistent entries found in table %s",
                len(inconsistent_ids),
                tab_name,
            )
            for bad_id in inconsistent_ids:
                logger.warning("  id: %s, name: %s", bad_id, db.get(bad_id).name)
        else:
            logger.info("'%s' table has no inconsistent entries.", tab_name)
        return inconsistent_ids

    def set_template(self, atoms: Atoms) -> None:
        """Check the size of provided cell and set as the currently active
        template in the settings object.

        Parameters:

            atoms (Atoms):
                Unrelaxed structure
        """
        self.settings.set_active_template(atoms=atoms)

    def _cf_name_exists(self, cf_name):
        """Return True if cluster name exists. Otherwise False.

        Parameters:

        cluster_name: str
            Cluster name to check
        """
        return cf_name in self.settings.all_cf_names

    def _confirm_cf_names_exists(self, cf_names):
        if not set(cf_names).issubset(self.settings.all_cf_names):
            raise ClusterNotTrackedError(
                "The correlation function of non-existing cluster is "
                "requested, but the name does not exist in "
                "ClusterExpansionSettings. Check that the cutoffs are "
                "correct, and try to run reconfigure_settings"
            )


def format_selection(select_cond=None, default_struct_type="initial"):
    """DB selection formatter. Will default to selecting
    all initial structures if None is specified."""
    select = []
    if select_cond is not None:
        for cond in select_cond:
            select.append(cond)
    else:
        select = [("struct_type", "=", default_struct_type)]
    return select



#!/usr/bin/env python
from pathlib import Path, PurePath
import click
from clease_cxx import has_parallel
import clease
from clease.version import __version__


@click.group()
@click.version_option(__version__)
def clease_cli():
    """The main CLEASE CLI"""


@clease_cli.command()
def info():
    """Print some information about the current CLEASE installation."""
    is_par = has_parallel()
    inst_path = get_install_path()
    s = [f"Version: {__version__}"]
    s += [f"C++ OpenMP: {is_par}"]
    s += [f"Install path: {inst_path}"]
    click.echo("\n".join(s))


def get_install_path() -> PurePath:
    """Get the location of the CLEASE installation"""
    return (Path(clease.__file__).parent).resolve()
